#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: ce_vrrp
version_added: "2.4"
short_description: Manages VRRP interfaces on HUAWEI CloudEngine devices.
description:
    - Manages VRRP interface attributes on HUAWEI CloudEngine devices.
author:
    - Li Yanfeng (@CloudEngine-Ansible)
options:
    interface:
        description:
            - Name of an interface. The value is a string of 1 to 63 characters.
    vrid:
        description:
            - VRRP backup group ID.
              The value is an integer ranging from 1 to 255.
        default: present
    virtual_ip :
        description:
            - Virtual IP address. The value is a string of 0 to 255 characters.
    vrrp_type:
        description:
            - Type of a VRRP backup group.
        choices: ['normal', 'member', 'admin']
    admin_ignore_if_down:
        description:
            - mVRRP ignores an interface Down event.
        type: bool
        default: 'no'
    admin_vrid:
        description:
            - Tracked mVRRP ID. The value is an integer ranging from 1 to 255.
    admin_interface:
        description:
            - Tracked mVRRP interface name. The value is a string of 1 to 63 characters.
    admin_flowdown:
        description:
            - Disable the flowdown function for service VRRP.
        type: bool
        default: 'no'
    priority:
        description:
            - Configured VRRP priority.
              The value ranges from 1 to 254. The default value is 100. A larger value indicates a higher priority.
    version:
        description:
            - VRRP version. The default version is v2.
        choices: ['v2','v3']
    advertise_interval:
        description:
            - Configured interval between sending advertisements, in milliseconds.
              Only the master router sends VRRP advertisements. The default value is 1000 milliseconds.
    preempt_timer_delay:
        description:
            - Preemption delay.
              The value is an integer ranging from 0 to 3600. The default value is 0.
    gratuitous_arp_interval:
        description:
            - Interval at which gratuitous ARP packets are sent, in seconds.
              The value ranges from 30 to 1200.The default value is 300.
    recover_delay:
        description:
            - Delay in recovering after an interface goes Up.
              The delay is used for interface flapping suppression.
              The value is an integer ranging from 0 to 3600.
              The default value is 0 seconds.
    holding_multiplier:
        description:
            - The configured holdMultiplier.The value is an integer ranging from 3 to 10. The default value is 3.
    auth_mode:
        description:
            - Authentication type used for VRRP packet exchanges between virtual routers.
              The values are noAuthentication, simpleTextPassword, md5Authentication.
              The default value is noAuthentication.
        choices: ['simple','md5','none']
    is_plain:
        description:
            - Select the display mode of an authentication key.
              By default, an authentication key is displayed in ciphertext.
        type: bool
        default: 'no'
    auth_key:
        description:
            - This object is set based on the authentication type.
              When noAuthentication is specified, the value is empty.
              When simpleTextPassword or md5Authentication is specified, the value is a string of 1 to 8 characters
              in plaintext and displayed as a blank text for security.
    fast_resume:
        description:
            - mVRRP's fast resume mode.
        choices: ['enable','disable']
    state:
        description:
            - Specify desired state of the resource.
        default: present
        choices: ['present','absent']

'''

EXAMPLES = '''
- name: vrrp module test
  hosts: cloudengine
  connection: local
  gather_facts: no
  vars:
    cli:
      host: "{{ inventory_hostname }}"
      port: "{{ ansible_ssh_port }}"
      username: "{{ username }}"
      password: "{{ password }}"
      transport: cli

  tasks:

  - name: Set vrrp version
    ce_vrrp:
      version: v3
      provider: "{{ cli }}"

  - name: Set vrrp gratuitous-arp interval
    ce_vrrp:
      gratuitous_arp_interval: 40
      mlag_id: 4
      provider: "{{ cli }}"

  - name: Set vrrp recover-delay
    ce_vrrp:
      recover_delay: 10
      provider: "{{ cli }}"

  - name: Set vrrp vrid virtual-ip
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      virtual_ip: 10.14.2.7
      provider: "{{ cli }}"

  - name: Set vrrp vrid admin
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      vrrp_type: admin
      provider: "{{ cli }}"

  - name: Set vrrp vrid fast_resume
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      fast_resume: enable
      provider: "{{ cli }}"

  - name: Set vrrp vrid holding-multiplier
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      holding_multiplier: 4
      provider: "{{ cli }}"

  - name: Set vrrp vrid preempt timer delay
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      preempt_timer_delay: 10
      provider: "{{ cli }}"

  - name: Set vrrp vrid admin-vrrp
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      admin_interface: 40GE2/0/9
      admin_vrid: 2
      vrrp_type: member
      provider: "{{ cli }}"

  - name: Set vrrp vrid authentication-mode
    ce_vrrp:
      interface: 40GE2/0/8
      vrid: 1
      is_plain: true
      auth_mode: simple
      auth_key: aaa
      provider: "{{ cli }}"
'''

RETURN = '''
changed:
    description: check to see if a change was made on the device
    returned: always
    type: boolean
    sample: true
proposed:
    description: k/v pairs of parameters passed into module
    returned: always
    type: dict
    sample: {
                "auth_key": "aaa",
                "auth_mode": "simple",
                "interface": "40GE2/0/8",
                "is_plain": true,
                "state": "present",
                "vrid": "1"
            }
existing:
    description: k/v pairs of existing aaa server
    returned: always
    type: dict
    sample: {
                "auth_mode": "none",
                "interface": "40GE2/0/8",
                "is_plain": "false",
                "vrid": "1",
                "vrrp_type": "normal"
            }
end_state:
    description: k/v pairs of aaa params after module execution
    returned: always
    type: dict
    sample: {
                "auth_mode": "simple",
                "interface": "40GE2/0/8",
                "is_plain": "true",
                "vrid": "1",
                "vrrp_type": "normal"
    }
updates:
    description: command sent to the device
    returned: always
    type: list
    sample: { "interface 40GE2/0/8",
              "vrrp vrid 1 authentication-mode simple plain aaa"}
'''

from xml.etree import ElementTree
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.cloudengine.ce import get_nc_config, set_nc_config, ce_argument_spec


CE_NC_GET_VRRP_GROUP_INFO = """
<filter type="subtree">
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGroups>
      <vrrpGroup>
        <ifName>%s</ifName>
        <vrrpId>%s</vrrpId>
      </vrrpGroup>
    </vrrpGroups>
  </vrrp>
</filter>
"""

CE_NC_SET_VRRP_GROUP_INFO_HEAD = """
<config>
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGroups>
      <vrrpGroup operation="merge">
        <ifName>%s</ifName>
        <vrrpId>%s</vrrpId>

"""
CE_NC_SET_VRRP_GROUP_INFO_TAIL = """
      </vrrpGroup>
    </vrrpGroups>
  </vrrp>
</config>

"""
CE_NC_GET_VRRP_GLOBAL_INFO = """
<filter type="subtree">
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGlobalCfg>
      <gratuitousArpFlag></gratuitousArpFlag>
      <gratuitousArpTimeOut></gratuitousArpTimeOut>
      <recoverDelay></recoverDelay>
      <version></version>
    </vrrpGlobalCfg>
  </vrrp>
</filter>

"""

CE_NC_SET_VRRP_GLOBAL_HEAD = """
<config>
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGlobalCfg operation="merge">
"""
CE_NC_SET_VRRP_GLOBAL_TAIL = """
    </vrrpGlobalCfg>
  </vrrp>
</config>

"""

CE_NC_GET_VRRP_VIRTUAL_IP_INFO = """
<filter type="subtree">
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGroups>
      <vrrpGroup>
        <vrrpId>%s</vrrpId>
        <ifName>%s</ifName>
        <virtualIps>
          <virtualIp>
            <virtualIpAddress></virtualIpAddress>
          </virtualIp>
        </virtualIps>
      </vrrpGroup>
    </vrrpGroups>
  </vrrp>
</filter>

"""
CE_NC_CREATE_VRRP_VIRTUAL_IP_INFO = """
<config>
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGroups>
      <vrrpGroup>
        <vrrpId>%s</vrrpId>
        <ifName>%s</ifName>
        <virtualIps>
          <virtualIp operation="create">
            <virtualIpAddress>%s</virtualIpAddress>
          </virtualIp>
        </virtualIps>
      </vrrpGroup>
    </vrrpGroups>
  </vrrp>
</config>

"""
CE_NC_DELETE_VRRP_VIRTUAL_IP_INFO = """
<config>
  <vrrp xmlns="http://www.huawei.com/netconf/vrp" content-version="1.0" format-version="1.0">
    <vrrpGroups>
      <vrrpGroup>
        <vrrpId>%s</vrrpId>
        <ifName>%s</ifName>
        <virtualIps>
          <virtualIp operation="delete">
            <virtualIpAddress>%s</virtualIpAddress>
          </virtualIp>
        </virtualIps>
      </vrrpGroup>
    </vrrpGroups>
  </vrrp>
</config>

"""


def is_valid_address(address):
    """check ip-address is valid"""

    if address.find('.') != -1:
        addr_list = address.split('.')
        if len(addr_list) != 4:
            return False
        for each_num in addr_list:
            if not each_num.isdigit():
                return False
            if int(each_num) > 255:
                return False
        return True

    return False


def get_interface_type(interface):
    """Gets the type of interface, such as 10GE, ETH-TRUNK, VLANIF..."""

    if interface is None:
        return None

    iftype = None

    if interface.upper().startswith('GE'):
        iftype = 'ge'
    elif interface.upper().startswith('10GE'):
        iftype = '10ge'
    elif interface.upper().startswith('25GE'):
        iftype = '25ge'
    elif interface.upper().startswith('40GE'):
        iftype = '40ge'
    elif interface.upper().startswith('100GE'):
        iftype = '100ge'
    elif interface.upper().startswith('ETH-TRUNK'):
        iftype = 'eth-trunk'
    elif interface.upper().startswith('NULL'):
        iftype = 'null'
    elif interface.upper().startswith('VLANIF'):
        iftype = 'vlanif'
    else:
        return None

    return iftype.lower()


class Vrrp(object):
    """
    Manages Manages vrrp information.
    """

    def __init__(self, argument_spec):
        self.spec = argument_spec
        self.module = None
        self.init_module()

        # module input info
        self.interface = self.module.params['interface']
        self.vrid = self.module.params['vrid']
        self.virtual_ip = self.module.params['virtual_ip']
        self.vrrp_type = self.module.params['vrrp_type']
        self.admin_ignore_if_down = self.module.params['admin_ignore_if_down']
        self.admin_vrid = self.module.params['admin_vrid']
        self.admin_interface = self.module.params['admin_interface']
        self.admin_flowdown = self.module.params['admin_flowdown']
        self.priority = self.module.params['priority']
        self.version = self.module.params['version']
        self.advertise_interval = self.module.params['advertise_interval']
        self.preempt_timer_delay = self.module.params['preempt_timer_delay']
        self.gratuitous_arp_interval = self.module.params[
            'gratuitous_arp_interval']
        self.recover_delay = self.module.params['recover_delay']
        self.holding_multiplier = self.module.params['holding_multiplier']
        self.auth_mode = self.module.params['auth_mode']
        self.is_plain = self.module.params['is_plain']
        self.auth_key = self.module.params['auth_key']
        self.fast_resume = self.module.params['fast_resume']
        self.state = self.module.params['state']

        # vrrp info
        self.vrrp_global_info = None
        self.virtual_ip_info = None
        self.vrrp_group_info = None

        # state
        self.changed = False
        self.updates_cmd = list()
        self.results = dict()
        self.existing = dict()
        self.proposed = dict()
        self.end_state = dict()

    def init_module(self):
        """ init module """

        self.module = AnsibleModule(
            argument_spec=self.spec, supports_check_mode=True)

    def get_virtual_ip_info(self):
        """ get vrrp virtual ip info."""
        virtual_ip_info = dict()
        conf_str = CE_NC_GET_VRRP_VIRTUAL_IP_INFO % (self.vrid, self.interface)
        xml_str = get_nc_config(self.module, conf_str)
        if "<data/>" in xml_str:
            return virtual_ip_info
        else:
            xml_str = xml_str.replace('\r', '').replace('\n', '').\
                replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
                replace('xmlns="http://www.huawei.com/netconf/vrp"', "")
            virtual_ip_info["vrrpVirtualIpInfos"] = list()
            root = ElementTree.fromstring(xml_str)
            vrrp_virtual_ip_infos = root.findall(
                "data/vrrp/vrrpGroups/vrrpGroup/virtualIps/virtualIp")
            if vrrp_virtual_ip_infos:
                for vrrp_virtual_ip_info in vrrp_virtual_ip_infos:
                    virtual_ip_dict = dict()
                    for ele in vrrp_virtual_ip_info:
                        if ele.tag in ["virtualIpAddress"]:
                            virtual_ip_dict[ele.tag] = ele.text
                    virtual_ip_info["vrrpVirtualIpInfos"].append(
                        virtual_ip_dict)
            return virtual_ip_info

    def get_vrrp_global_info(self):
        """ get vrrp global info."""

        vrrp_global_info = dict()
        conf_str = CE_NC_GET_VRRP_GLOBAL_INFO
        xml_str = get_nc_config(self.module, conf_str)
        if "<data/>" in xml_str:
            return vrrp_global_info
        else:
            xml_str = xml_str.replace('\r', '').replace('\n', '').\
                replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
                replace('xmlns="http://www.huawei.com/netconf/vrp"', "")

            root = ElementTree.fromstring(xml_str)
            global_info = root.findall(
                "data/vrrp/vrrpGlobalCfg")

            if global_info:
                for tmp in global_info:
                    for site in tmp:
                        if site.tag in ["gratuitousArpTimeOut", "gratuitousArpFlag", "recoverDelay", "version"]:
                            vrrp_global_info[site.tag] = site.text
            return vrrp_global_info

    def get_vrrp_group_info(self):
        """ get vrrp group info."""

        vrrp_group_info = dict()
        conf_str = CE_NC_GET_VRRP_GROUP_INFO % (self.interface, self.vrid)
        xml_str = get_nc_config(self.module, conf_str)
        if "<data/>" in xml_str:
            return vrrp_group_info
        else:
            xml_str = xml_str.replace('\r', '').replace('\n', '').\
                replace('xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"', "").\
                replace('xmlns="http://www.huawei.com/netconf/vrp"', "")

            root = ElementTree.fromstring(xml_str)
            global_info = root.findall(
                "data/vrrp/vrrpGroups/vrrpGroup")

            if global_info:
                for tmp in global_info:
                    for site in tmp:
                        if site.tag in ["ifName", "vrrpId", "priority", "advertiseInterval", "preemptMode", "delayTime",
                                        "authenticationMode", "authenticationKey", "vrrpType", "adminVrrpId",
                                        "adminIfName", "adminIgnoreIfDown", "isPlain", "unflowdown", "fastResume",
                                        "holdMultiplier"]:
                            vrrp_group_info[site.tag] = site.text
            return vrrp_group_info

    def check_params(self):
        """Check all input params"""

        # interface check
        if self.interface:
            intf_type = get_interface_type(self.interface)
            if not intf_type:
                self.module.fail_json(
                    msg='Error: Interface name of %s '
                        'is error.' % self.interface)

        # vrid check
        if self.vrid:
            if not self.vrid.isdigit():
                self.module.fail_json(
                    msg='Error: The value of vrid is an integer.')
            if int(self.vrid) < 1 or int(self.vrid) > 255:
                self.module.fail_json(
                    msg='Error: The value of vrid ranges from 1 to 255.')

        # virtual_ip check
        if self.virtual_ip:
            if not is_valid_address(self.virtual_ip):
                self.module.fail_json(
                    msg='Error: The %s is not a valid ip address.' % self.virtual_ip)

        # admin_vrid check
        if self.admin_vrid:
            if not self.admin_vrid.isdigit():
                self.module.fail_json(
                    msg='Error: The value of admin_vrid is an integer.')
            if int(self.admin_vrid) < 1 or int(self.admin_vrid) > 255:
                self.module.fail_json(
                    msg='Error: The value of admin_vrid ranges from 1 to 255.')

        # admin_interface check
        if self.admin_interface:
            intf_type = get_interface_type(self.admin_interface)
            if not intf_type:
                self.module.fail_json(
                    msg='Error: Admin interface name of %s '
                        'is error.' % self.admin_interface)

        # priority check
        if self.priority:
            if not self.priority.isdigit():
                self.module.fail_json(
                    msg='Error: The value of priority is an integer.')
            if int(self.priority) < 1 or int(self.priority) > 254:
                self.module.fail_json(
                    msg='Error: The value of priority ranges from 1 to 254. The default value is 100.')

        # advertise_interval check
        if self.advertise_interval:
            if not self.advertise_interval.isdigit():
                self.module.fail_json(
                    msg='Error: The value of advertise_interval is an integer.')
            if int(self.advertise_interval) < 1 or int(self.advertise_interval) > 255000:
                self.module.fail_json(
                    msg='Error: The value of advertise_interval ranges from 1 to 255000. The default value is 1000.')

        # preempt_timer_delay check
        if self.preempt_timer_delay:
            if not self.preempt_timer_delay.isdigit():
                self.module.fail_json(
                    msg='Error: The value of preempt_timer_delay is an integer.')
            if int(self.preempt_timer_delay) < 1 or int(self.preempt_timer_delay) > 3600:
                self.module.fail_json(
                    msg='Error: The value of preempt_timer_delay ranges from 1 to 3600. The default value is 0.')

        # holding_multiplier check
        if self.holding_multiplier:
            if not self.holding_multiplier.isdigit():
                self.module.fail_json(
                    msg='Error: The value of holding_multiplier is an integer.')
            if int(self.holding_multiplier) < 3 or int(self.holding_multiplier) > 10:
                self.module.fail_json(
                    msg='Error: The value of holding_multiplier ranges from 3 to 10. The default value is 3.')

        # auth_key check
        if self.auth_key:
            if len(self.auth_key) > 16 \
                    or len(self.auth_key.replace(' ', '')) < 1:
                self.module.fail_json(
                    msg='Error: The length of auth_key is not in the range from 1 to 16.')

    def is_virtual_ip_change(self):
        """whether virtual ip change"""

        if not self.virtual_ip_info:
            return True

        for info in self.virtual_ip_info["vrrpVirtualIpInfos"]:
            if info["virtualIpAddress"] == self.virtual_ip:
                return False
        return True

    def is_virtual_ip_exist(self):
        """whether virtual ip info exist"""

        if not self.virtual_ip_info:
            return False

        for info in self.virtual_ip_info["vrrpVirtualIpInfos"]:
            if info["virtualIpAddress"] == self.virtual_ip:
                return True
        return False

    def is_vrrp_global_info_change(self):
        """whether vrrp global attribute info change"""

        if not self.vrrp_global_info:
            return True

        if self.gratuitous_arp_interval:
            if self.vrrp_global_info["gratuitousArpFlag"] == "false":
                self.module.fail_json(msg="Error: gratuitousArpFlag is false.")
            if self.vrrp_global_info["gratuitousArpTimeOut"] != self.gratuitous_arp_interval:
                return True
        if self.recover_delay:
            if self.vrrp_global_info["recoverDelay"] != self.recover_delay:
                return True
        if self.version:
            if self.vrrp_global_info["version"] != self.version:
                return True
        return False

    def is_vrrp_global_info_exist(self):
        """whether vrrp global attribute info exist"""

        if self.gratuitous_arp_interval or self.recover_delay or self.version:
            if self.gratuitous_arp_interval:
                if self.vrrp_global_info["gratuitousArpFlag"] == "false":
                    self.module.fail_json(
                        msg="Error: gratuitousArpFlag is false.")
                if self.vrrp_global_info["gratuitousArpTimeOut"] != self.gratuitous_arp_interval:
                    return False
            if self.recover_delay:
                if self.vrrp_global_info["recoverDelay"] != self.recover_delay:
                    return False
            if self.version:
                if self.vrrp_global_info["version"] != self.version:
                    return False
            return True

        return False

    def is_vrrp_group_info_change(self):
        """whether vrrp group attribute info change"""

        if self.vrrp_type:
            if self.vrrp_group_info["vrrpType"] != self.vrrp_type:
                return True
        if self.admin_ignore_if_down:
            if self.vrrp_group_info["adminIgnoreIfDown"] != self.admin_ignore_if_down:
                return True
        if self.admin_vrid:
            if self.vrrp_group_info["adminVrrpId"] != self.admin_vrid:
                return True
        if self.admin_interface:
            if self.vrrp_group_info["adminIfName"] != self.admin_interface:
                return True
        if self.admin_flowdown:
            if self.vrrp_group_info["unflowdown"] != self.admin_flowdown:
                return True
        if self.priority:
            if self.vrrp_group_info["priority"] != self.priority:
                return True
        if self.fast_resume:
            fast_resume = "false"
            if self.fast_resume == "enable":
                fast_resume = "true"
            if self.vrrp_group_info["fastResume"] != fast_resume:
                return True
        if self.advertise_interval:
            if self.vrrp_group_info["advertiseInterval"] != self.advertise_interval:
                return True
        if self.preempt_timer_delay:
            if self.vrrp_group_info["delayTime"] != self.preempt_timer_delay:
                return True
        if self.holding_multiplier:
            if self.vrrp_group_info["holdMultiplier"] != self.holding_multiplier:
                return True
        if self.auth_mode:
            if self.vrrp_group_info["authenticationMode"] != self.auth_mode:
                return True
        if self.auth_key:
            return True
        if self.is_plain:
            if self.vrrp_group_info["isPlain"] != self.is_plain:
                return True

        return False

    def is_vrrp_group_info_exist(self):
        """whether vrrp group attribute info exist"""

        if self.vrrp_type:
            if self.vrrp_group_info["vrrpType"] != self.vrrp_type:
                return False
        if self.admin_ignore_if_down:
            if self.vrrp_group_info["adminIgnoreIfDown"] != self.admin_ignore_if_down:
                return False
        if self.admin_vrid:
            if self.vrrp_group_info["adminVrrpId"] != self.admin_vrid:
                return False
        if self.admin_interface:
            if self.vrrp_group_info["adminIfName"] != self.admin_interface:
                return False
        if self.admin_flowdown:
            if self.vrrp_group_info["unflowdown"] != self.admin_flowdown:
                return False
        if self.priority:
            if self.vrrp_group_info["priority"] != self.priority:
                return False
        if self.fast_resume:
            fast_resume = "false"
            if self.fast_resume == "enable":
                fast_resume = "true"
            if self.vrrp_group_info["fastResume"] != fast_resume:
                return False
        if self.advertise_interval:
            if self.vrrp_group_info["advertiseInterval"] != self.advertise_interval:
                return False
        if self.preempt_timer_delay:
            if self.vrrp_group_info["delayTime"] != self.preempt_timer_delay:
                return False
        if self.holding_multiplier:
            if self.vrrp_group_info["holdMultiplier"] != self.holding_multiplier:
                return False
        if self.auth_mode:
            if self.vrrp_group_info["authenticationMode"] != self.auth_mode:
                return False
        if self.is_plain:
            if self.vrrp_group_info["isPlain"] != self.is_plain:
                return False
        return True

    def create_virtual_ip(self):
        """create virtual ip info"""

        if self.is_virtual_ip_change():
            conf_str = CE_NC_CREATE_VRRP_VIRTUAL_IP_INFO % (
                self.vrid, self.interface, self.virtual_ip)
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: create virtual ip info failed.')

            self.updates_cmd.append("interface %s" % self.interface)
            self.updates_cmd.append(
                "vrrp vrid %s virtual-ip %s" % (self.vrid, self.virtual_ip))
            self.changed = True

    def delete_virtual_ip(self):
        """delete virtual ip info"""

        if self.is_virtual_ip_exist():
            conf_str = CE_NC_DELETE_VRRP_VIRTUAL_IP_INFO % (
                self.vrid, self.interface, self.virtual_ip)
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: delete virtual ip info failed.')

            self.updates_cmd.append("interface %s" % self.interface)
            self.updates_cmd.append(
                "undo vrrp vrid %s virtual-ip %s " % (self.vrid, self.virtual_ip))
            self.changed = True

    def set_vrrp_global(self):
        """set vrrp global attribute info"""

        if self.is_vrrp_global_info_change():
            conf_str = CE_NC_SET_VRRP_GLOBAL_HEAD
            if self.gratuitous_arp_interval:
                conf_str += "<gratuitousArpTimeOut>%s</gratuitousArpTimeOut>" % self.gratuitous_arp_interval
            if self.recover_delay:
                conf_str += "<recoverDelay>%s</recoverDelay>" % self.recover_delay
            if self.version:
                conf_str += "<version>%s</version>" % self.version
            conf_str += CE_NC_SET_VRRP_GLOBAL_TAIL
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: set vrrp global atrribute info failed.')

            if self.gratuitous_arp_interval:
                self.updates_cmd.append(
                    "vrrp gratuitous-arp interval %s" % self.gratuitous_arp_interval)

            if self.recover_delay:
                self.updates_cmd.append(
                    "vrrp recover-delay %s" % self.recover_delay)

            if self.version:
                version = "3"
                if self.version == "v2":
                    version = "2"
                self.updates_cmd.append("vrrp version %s" % version)
            self.changed = True

    def delete_vrrp_global(self):
        """delete vrrp global attribute info"""

        if self.is_vrrp_global_info_exist():
            conf_str = CE_NC_SET_VRRP_GLOBAL_HEAD
            if self.gratuitous_arp_interval:
                if self.gratuitous_arp_interval == "120":
                    self.module.fail_json(
                        msg='Error: The default value of gratuitous_arp_interval is 120.')
                gratuitous_arp_interval = "120"
                conf_str += "<gratuitousArpTimeOut>%s</gratuitousArpTimeOut>" % gratuitous_arp_interval
            if self.recover_delay:
                if self.recover_delay == "0":
                    self.module.fail_json(
                        msg='Error: The default value of recover_delay is 0.')
                recover_delay = "0"
                conf_str += "<recoverDelay>%s</recoverDelay>" % recover_delay
            if self.version:
                if self.version == "v2":
                    self.module.fail_json(
                        msg='Error: The default value of version is v2.')
                version = "v2"
                conf_str += "<version>%s</version>" % version
            conf_str += CE_NC_SET_VRRP_GLOBAL_TAIL
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: set vrrp global atrribute info failed.')
            if self.gratuitous_arp_interval:
                self.updates_cmd.append("undo vrrp gratuitous-arp interval")

            if self.recover_delay:
                self.updates_cmd.append("undo vrrp recover-delay")

            if self.version == "v3":
                self.updates_cmd.append("undo vrrp version")
            self.changed = True

    def set_vrrp_group(self):
        """set vrrp group attribute info"""

        if self.is_vrrp_group_info_change():
            conf_str = CE_NC_SET_VRRP_GROUP_INFO_HEAD % (
                self.interface, self.vrid)
            if self.vrrp_type:
                conf_str += "<vrrpType>%s</vrrpType>" % self.vrrp_type
            if self.admin_vrid:
                conf_str += "<adminVrrpId>%s</adminVrrpId>" % self.admin_vrid
            if self.admin_interface:
                conf_str += "<adminIfName>%s</adminIfName>" % self.admin_interface
                if self.admin_flowdown is True or self.admin_flowdown is False:
                    admin_flowdown = "false"
                    if self.admin_flowdown is True:
                        admin_flowdown = "true"
                    conf_str += "<unflowdown>%s</unflowdown>" % admin_flowdown
            if self.priority:
                conf_str += "<priority>%s</priority>" % self.priority
            if self.vrrp_type == "admin":
                if self.admin_ignore_if_down is True or self.admin_ignore_if_down is False:
                    admin_ignore_if_down = "false"
                    if self.admin_ignore_if_down is True:
                        admin_ignore_if_down = "true"
                    conf_str += "<adminIgnoreIfDown>%s</adminIgnoreIfDown>" % admin_ignore_if_down
            if self.fast_resume:
                fast_resume = "false"
                if self.fast_resume == "enable":
                    fast_resume = "true"
                conf_str += "<fastResume>%s</fastResume>" % fast_resume
            if self.advertise_interval:
                conf_str += "<advertiseInterval>%s</advertiseInterval>" % self.advertise_interval
            if self.preempt_timer_delay:
                conf_str += "<delayTime>%s</delayTime>" % self.preempt_timer_delay
            if self.holding_multiplier:
                conf_str += "<holdMultiplier>%s</holdMultiplier>" % self.holding_multiplier
            if self.auth_mode:
                conf_str += "<authenticationMode>%s</authenticationMode>" % self.auth_mode
            if self.auth_key:
                conf_str += "<authenticationKey>%s</authenticationKey>" % self.auth_key
            if self.auth_mode == "simple":
                is_plain = "false"
                if self.is_plain is True:
                    is_plain = "true"
                conf_str += "<isPlain>%s</isPlain>" % is_plain

            conf_str += CE_NC_SET_VRRP_GROUP_INFO_TAIL
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: set vrrp group atrribute info failed.')

            if self.interface and self.vrid:
                if self.vrrp_type == "admin":
                    if self.admin_ignore_if_down is True:
                        self.updates_cmd.append(
                            "interface %s" % self.interface)
                        self.updates_cmd.append(
                            "vrrp vrid %s admin ignore-if-down" % self.vrid)
                    else:
                        self.updates_cmd.append(
                            "interface %s" % self.interface)
                        self.updates_cmd.append(
                            "vrrp vrid %s admin" % self.vrid)

                if self.priority:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "vrrp vrid %s priority %s" % (self.vrid, self.priority))

                if self.fast_resume == "enable":
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "vrrp vrid %s fast-resume" % self.vrid)
                if self.fast_resume == "disable":
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s fast-resume" % self.vrid)

                if self.advertise_interval:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append("vrrp vrid %s timer advertise %s" % (
                        self.vrid, self.advertise_interval))

                if self.preempt_timer_delay:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append("vrrp vrid %s preempt timer delay %s" % (self.vrid,
                                                                                     self.preempt_timer_delay))

                if self.holding_multiplier:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "vrrp vrid %s holding-multiplier %s" % (self.vrid, self.holding_multiplier))

                if self.admin_vrid and self.admin_interface:
                    if self.admin_flowdown is True:
                        self.updates_cmd.append(
                            "interface %s" % self.interface)
                        self.updates_cmd.append("vrrp vrid %s track admin-vrrp interface %s vrid %s unflowdown" %
                                                (self.vrid, self.admin_interface, self.admin_vrid))
                    else:
                        self.updates_cmd.append(
                            "interface %s" % self.interface)
                        self.updates_cmd.append("vrrp vrid %s track admin-vrrp interface %s vrid %s" %
                                                (self.vrid, self.admin_interface, self.admin_vrid))

                if self.auth_mode and self.auth_key:
                    if self.auth_mode == "simple":
                        if self.is_plain is True:
                            self.updates_cmd.append(
                                "interface %s" % self.interface)
                            self.updates_cmd.append("vrrp vrid %s authentication-mode simple plain %s" %
                                                    (self.vrid, self.auth_key))
                        else:
                            self.updates_cmd.append(
                                "interface %s" % self.interface)
                            self.updates_cmd.append("vrrp vrid %s authentication-mode simple cipher %s" %
                                                    (self.vrid, self.auth_key))
                    if self.auth_mode == "md5":
                        self.updates_cmd.append(
                            "interface %s" % self.interface)
                        self.updates_cmd.append(
                            "vrrp vrid %s authentication-mode md5 %s" % (self.vrid, self.auth_key))
                self.changed = True

    def delete_vrrp_group(self):
        """delete vrrp group attribute info"""

        if self.is_vrrp_group_info_exist():
            conf_str = CE_NC_SET_VRRP_GROUP_INFO_HEAD % (
                self.interface, self.vrid)
            if self.vrrp_type:
                vrrp_type = self.vrrp_type
                if self.vrrp_type == "admin":
                    vrrp_type = "normal"
                if self.vrrp_type == "member" and self.admin_vrid and self.admin_interface:
                    vrrp_type = "normal"
                conf_str += "<vrrpType>%s</vrrpType>" % vrrp_type
            if self.priority:
                if self.priority == "100":
                    self.module.fail_json(
                        msg='Error: The default value of priority is 100.')
                priority = "100"
                conf_str += "<priority>%s</priority>" % priority

            if self.fast_resume:
                fast_resume = "false"
                if self.fast_resume == "enable":
                    fast_resume = "true"
                conf_str += "<fastResume>%s</fastResume>" % fast_resume
            if self.advertise_interval:
                if self.advertise_interval == "1000":
                    self.module.fail_json(
                        msg='Error: The default value of advertise_interval is 1000.')
                advertise_interval = "1000"
                conf_str += "<advertiseInterval>%s</advertiseInterval>" % advertise_interval
            if self.preempt_timer_delay:
                if self.preempt_timer_delay == "0":
                    self.module.fail_json(
                        msg='Error: The default value of preempt_timer_delay is 0.')
                preempt_timer_delay = "0"
                conf_str += "<delayTime>%s</delayTime>" % preempt_timer_delay
            if self.holding_multiplier:
                if self.holding_multiplier == "0":
                    self.module.fail_json(
                        msg='Error: The default value of holding_multiplier is 3.')
                holding_multiplier = "3"
                conf_str += "<holdMultiplier>%s</holdMultiplier>" % holding_multiplier
            if self.auth_mode:
                auth_mode = self.auth_mode
                if self.auth_mode == "md5" or self.auth_mode == "simple":
                    auth_mode = "none"
                conf_str += "<authenticationMode>%s</authenticationMode>" % auth_mode

            conf_str += CE_NC_SET_VRRP_GROUP_INFO_TAIL
            recv_xml = set_nc_config(self.module, conf_str)
            if "<ok/>" not in recv_xml:
                self.module.fail_json(
                    msg='Error: set vrrp global atrribute info failed.')
            if self.interface and self.vrid:
                if self.vrrp_type == "admin":
                    self.updates_cmd.append(
                        "undo vrrp vrid %s admin" % self.vrid)

                if self.priority:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s priority" % self.vrid)

                if self.fast_resume:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s fast-resume" % self.vrid)

                if self.advertise_interval:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s timer advertise" % self.vrid)

                if self.preempt_timer_delay:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s preempt timer delay" % self.vrid)

                if self.holding_multiplier:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s holding-multiplier" % self.vrid)

                if self.admin_vrid and self.admin_interface:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s track admin-vrrp" % self.vrid)

                if self.auth_mode:
                    self.updates_cmd.append("interface %s" % self.interface)
                    self.updates_cmd.append(
                        "undo vrrp vrid %s authentication-mode" % self.vrid)
                self.changed = True

    def get_proposed(self):
        """get proposed info"""

        if self.interface:
            self.proposed["interface"] = self.interface
        if self.vrid:
            self.proposed["vrid"] = self.vrid
        if self.virtual_ip:
            self.proposed["virtual_ip"] = self.virtual_ip
        if self.vrrp_type:
            self.proposed["vrrp_type"] = self.vrrp_type
        if self.admin_vrid:
            self.proposed["admin_vrid"] = self.admin_vrid
        if self.admin_interface:
            self.proposed["admin_interface"] = self.admin_interface
        if self.admin_flowdown:
            self.proposed["unflowdown"] = self.admin_flowdown
        if self.admin_ignore_if_down:
            self.proposed["admin_ignore_if_down"] = self.admin_ignore_if_down
        if self.priority:
            self.proposed["priority"] = self.priority
        if self.version:
            self.proposed["version"] = self.version
        if self.advertise_interval:
            self.proposed["advertise_interval"] = self.advertise_interval
        if self.preempt_timer_delay:
            self.proposed["preempt_timer_delay"] = self.preempt_timer_delay
        if self.gratuitous_arp_interval:
            self.proposed[
                "gratuitous_arp_interval"] = self.gratuitous_arp_interval
        if self.recover_delay:
            self.proposed["recover_delay"] = self.recover_delay
        if self.holding_multiplier:
            self.proposed["holding_multiplier"] = self.holding_multiplier
        if self.auth_mode:
            self.proposed["auth_mode"] = self.auth_mode
        if self.is_plain:
            self.proposed["is_plain"] = self.is_plain
        if self.auth_key:
            self.proposed["auth_key"] = self.auth_key
        if self.fast_resume:
            self.proposed["fast_resume"] = self.fast_resume
        if self.state:
            self.proposed["state"] = self.state

    def get_existing(self):
        """get existing info"""

        if self.gratuitous_arp_interval:
            self.existing["gratuitous_arp_interval"] = self.vrrp_global_info[
                "gratuitousArpTimeOut"]
        if self.version:
            self.existing["version"] = self.vrrp_global_info["version"]
        if self.recover_delay:
            self.existing["recover_delay"] = self.vrrp_global_info[
                "recoverDelay"]

        if self.virtual_ip:
            if self.virtual_ip_info:
                self.existing["interface"] = self.interface
                self.existing["vrid"] = self.vrid
                self.existing["virtual_ip_info"] = self.virtual_ip_info[
                    "vrrpVirtualIpInfos"]

        if self.vrrp_group_info:
            self.existing["interface"] = self.vrrp_group_info["ifName"]
            self.existing["vrid"] = self.vrrp_group_info["vrrpId"]
            self.existing["vrrp_type"] = self.vrrp_group_info["vrrpType"]
            if self.vrrp_type == "admin":
                self.existing["admin_ignore_if_down"] = self.vrrp_group_info[
                    "authenticationMode"]
            if self.admin_vrid and self.admin_interface:
                self.existing["admin_vrid"] = self.vrrp_group_info[
                    "adminVrrpId"]
                self.existing["admin_interface"] = self.vrrp_group_info[
                    "adminIfName"]
                self.existing["admin_flowdown"] = self.vrrp_group_info[
                    "unflowdown"]
            if self.priority:
                self.existing["priority"] = self.vrrp_group_info["priority"]
            if self.advertise_interval:
                self.existing["advertise_interval"] = self.vrrp_group_info[
                    "advertiseInterval"]
            if self.preempt_timer_delay:
                self.existing["preempt_timer_delay"] = self.vrrp_group_info[
                    "delayTime"]
            if self.holding_multiplier:
                self.existing["holding_multiplier"] = self.vrrp_group_info[
                    "holdMultiplier"]
            if self.fast_resume:
                fast_resume_exist = "disable"
                fast_resume = self.vrrp_group_info["fastResume"]
                if fast_resume == "true":
                    fast_resume_exist = "enable"
                self.existing["fast_resume"] = fast_resume_exist
            if self.auth_mode:
                self.existing["auth_mode"] = self.vrrp_group_info[
                    "authenticationMode"]
                self.existing["is_plain"] = self.vrrp_group_info["isPlain"]

    def get_end_state(self):
        """get end state info"""

        if self.gratuitous_arp_interval or self.version or self.recover_delay:
            self.vrrp_global_info = self.get_vrrp_global_info()
        if self.interface and self.vrid:
            if self.virtual_ip:
                self.virtual_ip_info = self.get_virtual_ip_info()
            if self.virtual_ip_info:
                self.vrrp_group_info = self.get_vrrp_group_info()

        if self.gratuitous_arp_interval:
            self.end_state["gratuitous_arp_interval"] = self.vrrp_global_info[
                "gratuitousArpTimeOut"]
        if self.version:
            self.end_state["version"] = self.vrrp_global_info["version"]
        if self.recover_delay:
            self.end_state["recover_delay"] = self.vrrp_global_info[
                "recoverDelay"]

        if self.virtual_ip:
            if self.virtual_ip_info:
                self.end_state["interface"] = self.interface
                self.end_state["vrid"] = self.vrid
                self.end_state["virtual_ip_info"] = self.virtual_ip_info[
                    "vrrpVirtualIpInfos"]

        if self.vrrp_group_info:
            self.end_state["interface"] = self.vrrp_group_info["ifName"]
            self.end_state["vrid"] = self.vrrp_group_info["vrrpId"]
            self.end_state["vrrp_type"] = self.vrrp_group_info["vrrpType"]
            if self.vrrp_type == "admin":
                self.end_state["admin_ignore_if_down"] = self.vrrp_group_info[
                    "authenticationMode"]
            if self.admin_vrid and self.admin_interface:
                self.existing["admin_vrid"] = self.vrrp_group_info[
                    "adminVrrpId"]
                self.end_state["admin_interface"] = self.vrrp_group_info[
                    "adminIfName"]
                self.end_state["admin_flowdown"] = self.vrrp_group_info[
                    "unflowdown"]
            if self.priority:
                self.end_state["priority"] = self.vrrp_group_info["priority"]
            if self.advertise_interval:
                self.end_state["advertise_interval"] = self.vrrp_group_info[
                    "advertiseInterval"]
            if self.preempt_timer_delay:
                self.end_state["preempt_timer_delay"] = self.vrrp_group_info[
                    "delayTime"]
            if self.holding_multiplier:
                self.end_state["holding_multiplier"] = self.vrrp_group_info[
                    "holdMultiplier"]
            if self.fast_resume:
                fast_resume_end = "disable"
                fast_resume = self.vrrp_group_info["fastResume"]
                if fast_resume == "true":
                    fast_resume_end = "enable"
                self.end_state["fast_resume"] = fast_resume_end
            if self.auth_mode:
                self.end_state["auth_mode"] = self.vrrp_group_info[
                    "authenticationMode"]
                self.end_state["is_plain"] = self.vrrp_group_info["isPlain"]

    def work(self):
        """worker"""

        self.check_params()
        if self.gratuitous_arp_interval or self.version or self.recover_delay:
            self.vrrp_global_info = self.get_vrrp_global_info()
        if self.interface and self.vrid:
            self.virtual_ip_info = self.get_virtual_ip_info()
            if self.virtual_ip_info:
                self.vrrp_group_info = self.get_vrrp_group_info()
        self.get_proposed()
        self.get_existing()

        if self.gratuitous_arp_interval or self.version or self.recover_delay:
            if self.state == "present":
                self.set_vrrp_global()
            else:
                self.delete_vrrp_global()
        else:
            if not self.interface or not self.vrid:
                self.module.fail_json(
                    msg='Error: interface, vrid must be config at the same time.')

        if self.interface and self.vrid:
            if self.virtual_ip:
                if self.state == "present":
                    self.create_virtual_ip()
                else:
                    self.delete_virtual_ip()
            else:
                if not self.vrrp_group_info:
                    self.module.fail_json(
                        msg='Error: The VRRP group does not exist.')
                if self.admin_ignore_if_down is True:
                    if self.vrrp_type != "admin":
                        self.module.fail_json(
                            msg='Error: vrrpType must be admin when admin_ignore_if_down is true.')
                if self.admin_interface or self.admin_vrid:
                    if self.vrrp_type != "member":
                        self.module.fail_json(
                            msg='Error: it binds a VRRP group to an mVRRP group, vrrp_type must be "member".')
                    if not self.vrrp_type or not self.interface or not self.vrid:
                        self.module.fail_json(
                            msg='Error: admin_interface admin_vrid vrrp_type interface vrid must '
                                'be config at the same time.')
                if self.auth_mode == "md5" and self.is_plain is True:
                    self.module.fail_json(
                        msg='Error: is_plain can not be True when auth_mode is md5.')

                if self.state == "present":
                    self.set_vrrp_group()
                else:
                    self.delete_vrrp_group()

        self.get_end_state()
        self.results['changed'] = self.changed
        self.results['proposed'] = self.proposed
        self.results['existing'] = self.existing
        self.results['end_state'] = self.end_state
        if self.changed:
            self.results['updates'] = self.updates_cmd
        else:
            self.results['updates'] = list()

        self.module.exit_json(**self.results)


def main():
    """ Module main """

    argument_spec = dict(
        interface=dict(type='str'),
        vrid=dict(type='str'),
        virtual_ip=dict(type='str'),
        vrrp_type=dict(type='str', choices=['normal', 'member', 'admin']),
        admin_ignore_if_down=dict(type='bool', default=False),
        admin_vrid=dict(type='str'),
        admin_interface=dict(type='str'),
        admin_flowdown=dict(type='bool', default=False),
        priority=dict(type='str'),
        version=dict(type='str', choices=['v2', 'v3']),
        advertise_interval=dict(type='str'),
        preempt_timer_delay=dict(type='str'),
        gratuitous_arp_interval=dict(type='str'),
        recover_delay=dict(type='str'),
        holding_multiplier=dict(type='str'),
        auth_mode=dict(type='str', choices=['simple', 'md5', 'none']),
        is_plain=dict(type='bool', default=False),
        auth_key=dict(type='str'),
        fast_resume=dict(type='str', choices=['enable', 'disable']),
        state=dict(type='str', default='present',
                   choices=['present', 'absent'])
    )

    argument_spec.update(ce_argument_spec)
    module = Vrrp(argument_spec=argument_spec)
    module.work()


if __name__ == '__main__':
    main()



import json
import sys

from ansible.module_utils.urls import fetch_url


class Response(object):

    def __init__(self, resp, info):
        self.body = None
        if resp:
            self.body = resp.read()
        self.info = info

    @property
    def json(self):
        if not self.body:
            if "body" in self.info:
                return json.loads(self.info["body"])
            return None
        try:
            return json.loads(self.body)
        except ValueError:
            return None

    @property
    def status_code(self):
        return self.info["status"]

    @property
    def ok(self):
        return self.status_code in (200, 201, 202, 204)


class ScalewayAPI(object):

    def __init__(self, module, base_url, headers=None):
        self.module = module
        self.headers = {'User-Agent': self.get_user_agent_string(module),
                        'Content-type': 'application/json'}
        if headers is not None:
            self.headers.update(headers)
        self.base_url = base_url

    def _url_builder(self, path):
        if path[0] == '/':
            path = path[1:]
        return '%s/%s' % (self.base_url, path)

    def send(self, method, path, data=None, headers=None):
        url = self._url_builder(path)
        data = self.module.jsonify(data)
        timeout = self.module.params['timeout']

        if headers is not None:
            self.headers.update(headers)

        resp, info = fetch_url(self.module, url, data=data, headers=self.headers, method=method, timeout=timeout)

        # Exceptions in fetch_url may result in a status -1, the ensures a proper error to the user in all cases
        if info['status'] == -1:
            self.module.fail_json(msg=info['msg'])

        return Response(resp, info)

    @staticmethod
    def get_user_agent_string(module):
        return "ansible %s Python %s" % (module.ansible_version, sys.version.split(' ')[0])

    def get(self, path, data=None, headers=None):
        return self.send('GET', path, data, headers)

    def put(self, path, data=None, headers=None):
        return self.send('PUT', path, data, headers)

    def post(self, path, data=None, headers=None):
        return self.send('POST', path, data, headers)

    def delete(self, path, data=None, headers=None):
        return self.send('DELETE', path, data, headers)

    def patch(self, path, data=None, headers=None):
        return self.send("PATCH", path, data, headers)

    def update(self, path, data=None, headers=None):
        return self.send("UPDATE", path, data, headers)


SCALEWAY_LOCATION = {
    'par1': {'name': 'Paris 1', 'country': 'FR', "api_endpoint": 'https://cp-par1.scaleway.com'},
    'EMEA-FR-PAR1': {'name': 'Paris 1', 'country': 'FR', "api_endpoint": 'https://cp-par1.scaleway.com'},

    'ams1': {'name': 'Amsterdam 1', 'country': 'NL', "api_endpoint": 'https://cp-ams1.scaleway.com'},
    'EMEA-NL-EVS': {'name': 'Amsterdam 1', 'country': 'NL', "api_endpoint": 'https://cp-ams1.scaleway.com'}
}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Liran Nisanov <lirannis@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# this is a windows documentation stub.  actual code lives in the .ps1
# file of the same name

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: win_pagefile
version_added: "2.4"
short_description: Query or change pagefile configuration
description:
    - Query current pagefile configuration.
    - Enable/Disable AutomaticManagedPagefile.
    - Create new or override pagefile configuration.
options:
  drive:
    description:
      - The drive of the pagefile.
  initial_size:
    description:
      - The initial size of the pagefile in megabytes.
  maximum_size:
    description:
      - The maximum size of the pagefile in megabytes.
  override:
    description:
      - Override the current pagefile on the drive.
    type: bool
    default: 'yes'
  system_managed:
    description:
      - Configures current pagefile to be managed by the system.
    type: bool
    default: 'no'
  automatic:
    description:
      - Configures AutomaticManagedPagefile for the entire system.
    type: bool
  remove_all:
    description:
      - Remove all pagefiles in the system, not including automatic managed.
    type: bool
    default: 'no'
  test_path:
    description:
      - Use Test-Path on the drive to make sure the drive is accessible before creating the pagefile.
    type: bool
    default: 'yes'
  state:
    description:
      - State of the pagefile.
    choices: [ absent, present, query ]
    default: query
notes:
- There is difference between automatic managed pagefiles that configured once for the entire system and system managed pagefile that configured per pagefile.
- InitialSize 0 and MaximumSize 0 means the pagefile is managed by the system.
- Value out of range exception may be caused by several different issues, two common problems - No such drive, Pagefile size is too small.
- Setting a pagefile when AutomaticManagedPagefile is on will disable the AutomaticManagedPagefile.
author:
- Liran Nisanov (@LiranNis)
'''

EXAMPLES = r'''
- name: Query pagefiles configuration
  win_pagefile:

- name: Query C pagefile
  win_pagefile:
    drive: C

- name: Set C pagefile, don't override if exists
  win_pagefile:
    drive: C
    initial_size: 1024
    maximum_size: 1024
    override: no
    state: present

- name: Set C pagefile, override if exists
  win_pagefile:
    drive: C
    initial_size: 1024
    maximum_size: 1024
    state: present

- name: Remove C pagefile
  win_pagefile:
    drive: C
    state: absent

- name: Remove all current pagefiles, enable AutomaticManagedPagefile and query at the end
  win_pagefile:
    remove_all: yes
    automatic: yes

- name: Remove all pagefiles disable AutomaticManagedPagefile and set C pagefile
  win_pagefile:
    drive: C
    initial_size: 2048
    maximum_size: 2048
    remove_all: yes
    automatic: no
    state: present

- name: Set D pagefile, override if exists
  win_pagefile:
    drive: d
    initial_size: 1024
    maximum_size: 1024
    state: present
'''

RETURN = r'''
automatic_managed_pagefiles:
    description: Whether the pagefiles is automatically managed.
    returned: When state is query.
    type: boolean
    sample: true
pagefiles:
    description: Contains caption, description, initial_size, maximum_size and name for each pagefile in the system.
    returned: When state is query.
    type: list
    sample:
      [{"caption": "c:\\ 'pagefile.sys'", "description": "'pagefile.sys' @ c:\\", "initial_size": 2048, "maximum_size": 2048, "name": "c:\\pagefile.sys"},
       {"caption": "d:\\ 'pagefile.sys'", "description": "'pagefile.sys' @ d:\\", "initial_size": 1024, "maximum_size": 1024, "name": "d:\\pagefile.sys"}]

'''



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Henrik Wallström <henrik@wallstroms.nu>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: win_iis_webapppool
version_added: "2.0"
short_description: Configure IIS Web Application Pools
description:
  - Creates, removes and configures an IIS Web Application Pool.
options:
  attributes:
    description:
      - This field is a free form dictionary value for the application pool
        attributes.
      - These attributes are based on the naming standard at
        U(https://www.iis.net/configreference/system.applicationhost/applicationpools/add#005),
        see the examples section for more details on how to set this.
      - You can also set the attributes of child elements like cpu and
        processModel, see the examples to see how it is done.
      - While you can use the numeric values for enums it is recommended to use
        the enum name itself, e.g. use SpecificUser instead of 3 for
        processModel.identityType.
      - managedPipelineMode may be either "Integrated" or "Classic".
      - startMode may be either "OnDemand" or "AlwaysRunning".
      - Use C(state) module parameter to modify the state of the app pool.
      - When trying to set 'processModel.password' and you receive a 'Value
        does fall within the expected range' error, you have a corrupted
        keystore. Please follow
        U(http://structuredsight.com/2014/10/26/im-out-of-range-youre-out-of-range/)
        to help fix your host.
  name:
    description:
      - Name of the application pool.
    required: yes
  state:
    choices: [ absent, present, restarted, started, stopped ]
    default: present
    description:
      - The state of the application pool.
      - If C(absent) will ensure the app pool is removed.
      - If C(present) will ensure the app pool is configured and exists.
      - If C(restarted) will ensure the app pool exists and will restart, this
        is never idempotent.
      - If C(started) will ensure the app pool exists and is started.
      - If C(stopped) will ensure the app pool exists and is stopped.
author:
- Henrik Wallström (@henrikwallstrom)
- Jordan Borean (@jborean93)
'''

EXAMPLES = r'''
- name: return information about an existing application pool
  win_iis_webapppool:
    name: DefaultAppPool
    state: present

- name: create a new application pool in 'Started' state
  win_iis_webapppool:
    name: AppPool
    state: started

- name: stop an application pool
  win_iis_webapppool:
    name: AppPool
    state: stopped

- name: restart an application pool (non-idempotent)
  win_iis_webapppool:
    name: AppPool
    state: restart

- name: change application pool attributes using new dict style
  win_iis_webapppool:
    name: AppPool
    attributes:
      managedRuntimeVersion: v4.0
      autoStart: no

- name: creates an application pool, sets attributes and starts it
  win_iis_webapppool:
    name: AnotherAppPool
    state: started
    attributes:
      managedRuntimeVersion: v4.0
      autoStart: no

# In the below example we are setting attributes in child element processModel
# https://www.iis.net/configreference/system.applicationhost/applicationpools/add/processmodel
- name: manage child element and set identity of application pool
  win_iis_webapppool:
    name: IdentitiyAppPool
    state: started
    attributes:
      managedPipelineMode: Classic
      processModel.identityType: SpecificUser
      processModel.userName: '{{ansible_user}}'
      processModel.password: '{{ansible_password}}'
      processModel.loadUserProfile: True

- name: manage a timespan attribute
  win_iis_webapppool:
    name: TimespanAppPool
    state: started
    attributes:
      # Timespan with full string "day:hour:minute:second.millisecond"
      recycling.periodicRestart.time: "00:00:05:00.000000"
      recycling.periodicRestart.schedule: ["00:10:00", "05:30:00"]
      # Shortened timespan "hour:minute:second"
      processModel.pingResponseTime: "00:03:00"
'''

RETURN = r'''
attributes:
  description: Application Pool attributes that were set and processed by this
    module invocation.
  returned: success
  type: dictionary
  sample:
    enable32BitAppOnWin64: "true"
    managedRuntimeVersion: "v4.0"
    managedPipelineMode: "Classic"
info:
  description: Information on current state of the Application Pool. See
    https://www.iis.net/configreference/system.applicationhost/applicationpools/add#005
    for the full list of return attributes based on your IIS version.
  returned: success
  type: complex
  sample:
  contains:
    attributes:
      description: Key value pairs showing the current Application Pool attributes.
      returned: success
      type: dictionary
      sample:
        autoStart: true
        managedRuntimeLoader: "webengine4.dll"
        managedPipelineMode: "Classic"
        name: "DefaultAppPool"
        CLRConfigFile: ""
        passAnonymousToken: true
        applicationPoolSid: "S-1-5-82-1352790163-598702362-1775843902-1923651883-1762956711"
        queueLength: 1000
        managedRuntimeVersion: "v4.0"
        state: "Started"
        enableConfigurationOverride: true
        startMode: "OnDemand"
        enable32BitAppOnWin64: true
    cpu:
      description: Key value pairs showing the current Application Pool cpu attributes.
      returned: success
      type: dictionary
      sample:
        action: "NoAction"
        limit: 0
        resetInterval:
          Days: 0
          Hours: 0
    failure:
      description: Key value pairs showing the current Application Pool failure attributes.
      returned: success
      type: dictionary
      sample:
        autoShutdownExe: ""
        orphanActionExe: ""
        rapidFailProtextionInterval:
          Days: 0
          Hours: 0
    name:
      description: Name of Application Pool that was processed by this module invocation.
      returned: success
      type: string
      sample: "DefaultAppPool"
    processModel:
      description: Key value pairs showing the current Application Pool processModel attributes.
      returned: success
      type: dictionary
      sample:
        identityType: "ApplicationPoolIdentity"
        logonType: "LogonBatch"
        pingInterval:
          Days: 0
          Hours: 0
    recycling:
      description: Key value pairs showing the current Application Pool recycling attributes.
      returned: success
      type: dictionary
      sample:
        disallowOverlappingRotation: false
        disallowRotationOnConfigChange: false
        logEventOnRecycle: "Time,Requests,Schedule,Memory,IsapiUnhealthy,OnDemand,ConfigChange,PrivateMemory"
    state:
      description: Current runtime state of the pool as the module completed.
      returned: success
      type: string
      sample: "Started"
'''



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2017, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'network'}


DOCUMENTATION = """
---
module: junos_netconf
version_added: "2.1"
author: "Peter Sprygada (@privateip)"
short_description: Configures the Junos Netconf system service
description:
  - This module provides an abstraction that enables and configures
    the netconf system service running on Junos devices.  This module
    can be used to easily enable the Netconf API. Netconf provides
    a programmatic interface for working with configuration and state
    resources as defined in RFC 6242.
extends_documentation_fragment: junos
options:
  netconf_port:
    description:
      - This argument specifies the port the netconf service should
        listen on for SSH connections.  The default port as defined
        in RFC 6242 is 830.
    required: false
    default: 830
    aliases: ['listens_on']
    version_added: "2.2"
  state:
    description:
      - Specifies the state of the C(junos_netconf) resource on
        the remote device.  If the I(state) argument is set to
        I(present) the netconf service will be configured.  If the
        I(state) argument is set to I(absent) the netconf service
        will be removed from the configuration.
    required: false
    default: present
    choices: ['present', 'absent']
notes:
  - Tested against vSRX JUNOS version 15.1X49-D15.4, vqfx-10000 JUNOS Version 15.1X53-D60.4.
  - Recommended connection is C(network_cli). See L(the Junos OS Platform Options,../network/user_guide/platform_junos.html).
  - This module also works with C(local) connections for legacy playbooks.
"""

EXAMPLES = """
- name: enable netconf service on port 830
  junos_netconf:
    listens_on: 830
    state: present

- name: disable netconf service
  junos_netconf:
    state: absent
"""

RETURN = """
commands:
  description: Returns the command sent to the remote device
  returned: when changed is True
  type: str
  sample: 'set system services netconf ssh port 830'
"""
import re

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.network.junos.junos import junos_argument_spec, get_connection
from ansible.module_utils.network.junos.junos import commit_configuration, discard_changes
from ansible.module_utils.network.common.utils import to_list
from ansible.module_utils.six import iteritems

USE_PERSISTENT_CONNECTION = True


def map_obj_to_commands(updates, module):
    want, have = updates
    commands = list()

    if want['state'] == 'absent':
        if have['state'] == 'present':
            commands.append('delete system services netconf')
    else:
        if have['state'] == 'absent' or want['netconf_port'] != have.get('netconf_port'):
            commands.append(
                'set system services netconf ssh port %s' % want['netconf_port']
            )

    return commands


def parse_port(config):
    match = re.search(r'port (\d+)', config)
    if match:
        return int(match.group(1))


def map_config_to_obj(module):
    conn = get_connection(module)
    out = conn.get(command='show configuration system services netconf')
    if out is None:
        module.fail_json(msg='unable to retrieve current config')
    config = str(out).strip()

    obj = {'state': 'absent'}
    if 'ssh' in config:
        obj.update({
            'state': 'present',
            'netconf_port': parse_port(config)
        })
    return obj


def validate_netconf_port(value, module):
    if not 1 <= value <= 65535:
        module.fail_json(msg='netconf_port must be between 1 and 65535')


def map_params_to_obj(module):
    obj = {
        'netconf_port': module.params['netconf_port'],
        'state': module.params['state']
    }

    for key, value in iteritems(obj):
        # validate the param value (if validator func exists)
        validator = globals().get('validate_%s' % key)
        if callable(validator):
            validator(value, module)

    return obj


def load_config(module, config, commit=False):
    conn = get_connection(module)

    conn.edit_config(to_list(config) + ['top'])
    diff = conn.compare_configuration()
    if diff:
        if commit:
            commit_configuration(module)

        else:
            discard_changes(module)

    return str(diff).strip()


def main():
    """main entry point for module execution
    """
    argument_spec = dict(
        netconf_port=dict(type='int', default=830, aliases=['listens_on']),
        state=dict(default='present', choices=['present', 'absent']),
    )

    argument_spec.update(junos_argument_spec)

    module = AnsibleModule(argument_spec=argument_spec,
                           supports_check_mode=True)

    warnings = list()
    result = {'changed': False, 'warnings': warnings}

    want = map_params_to_obj(module)
    have = map_config_to_obj(module)

    commands = map_obj_to_commands((want, have), module)
    result['commands'] = commands

    if commands:
        commit = not module.check_mode
        diff = load_config(module, commands, commit=commit)
        if diff:
            if module._diff:
                result['diff'] = {'prepared': diff}
            result['changed'] = True

    module.exit_json(**result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: aci_epg_to_contract
short_description: Bind EPGs to Contracts (fv:RsCons, fv:RsProv)
description:
- Bind EPGs to Contracts on Cisco ACI fabrics.
notes:
- The C(tenant), C(app_profile), C(EPG), and C(Contract) used must exist before using this module in your playbook.
  The M(aci_tenant), M(aci_ap), M(aci_epg), and M(aci_contract) modules can be used for this.
- More information about the internal APIC classes B(fv:RsCons) and B(fv:RsProv) from
  L(the APIC Management Information Model reference,https://developer.cisco.com/docs/apic-mim-ref/).
author:
- Jacob McGill (@jmcgill298)
version_added: '2.4'
options:
  ap:
    description:
    - Name of an existing application network profile, that will contain the EPGs.
    aliases: [ app_profile, app_profile_name ]
  contract:
    description:
    - The name of the contract.
    aliases: [ contract_name ]
  contract_type:
    description:
    - Determines if the EPG should Provide or Consume the Contract.
    required: yes
    choices: [ consumer, provider ]
  epg:
    description:
    - The name of the end point group.
    aliases: [ epg_name ]
  priority:
    description:
    - QoS class.
    - The APIC defaults to C(unspecified) when unset during creation.
    choices: [ level1, level2, level3, unspecified ]
  provider_match:
    description:
    - The matching algorithm for Provided Contracts.
    - The APIC defaults to C(at_least_one) when unset during creation.
    choices: [ all, at_least_one, at_most_one, none ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    choices: [ absent, present, query ]
    default: present
  tenant:
    description:
    - Name of an existing tenant.
    aliases: [ tenant_name ]
extends_documentation_fragment: aci
'''

EXAMPLES = r'''
- name: Add a new contract to EPG binding
  aci_epg_to_contract:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: anstest
    ap: anstest
    epg: anstest
    contract: anstest_http
    contract_type: provider
    state: present

- name: Remove an existing contract to EPG binding
  aci_epg_to_contract:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: anstest
    ap: anstest
    epg: anstest
    contract: anstest_http
    contract_type: provider
    state: absent

- name: Query a specific contract to EPG binding
  aci_epg_to_contract:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: anstest
    ap: anstest
    epg: anstest
    contract: anstest_http
    contract_type: provider
    state: query

- name: Query all provider contract to EPG bindings
  aci_epg_to_contract:
    host: apic
    username: admin
    password: SomeSecretPassword
    contract_type: provider
    state: query
'''

RETURN = r'''
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: string
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: string
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: string
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: string
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: string
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
'''

from ansible.module_utils.network.aci.aci import ACIModule, aci_argument_spec
from ansible.module_utils.basic import AnsibleModule

ACI_CLASS_MAPPING = {"consumer": {"class": "fvRsCons", "rn": "rscons-"}, "provider": {"class": "fvRsProv", "rn": "rsprov-"}}
PROVIDER_MATCH_MAPPING = {"all": "All", "at_least_one": "AtleastOne", "at_most_one": "AtmostOne", "none": "None"}


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(
        ap=dict(type='str', aliases=['app_profile', 'app_profile_name']),  # Not required for querying all objects
        epg=dict(type='str', aliases=['epg_name']),  # Not required for querying all objects
        contract=dict(type='str', aliases=['contract_name']),  # Not required for querying all objects
        contract_type=dict(type='str', required=True, choices=['consumer', 'provider']),
        priority=dict(type='str', choices=['level1', 'level2', 'level3', 'unspecified']),
        provider_match=dict(type='str', choices=['all', 'at_least_one', 'at_most_one', 'none']),
        state=dict(type='str', default='present', choices=['absent', 'present', 'query']),
        tenant=dict(type='str', aliases=['tenant_name']),  # Not required for querying all objects
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ['state', 'absent', ['ap', 'contract', 'epg', 'tenant']],
            ['state', 'present', ['ap', 'contract', 'epg', 'tenant']],
        ],
    )

    ap = module.params['ap']
    contract = module.params['contract']
    contract_type = module.params['contract_type']
    epg = module.params['epg']
    priority = module.params['priority']
    provider_match = module.params['provider_match']
    if provider_match is not None:
        provider_match = PROVIDER_MATCH_MAPPING[provider_match]
    state = module.params['state']
    tenant = module.params['tenant']

    aci_class = ACI_CLASS_MAPPING[contract_type]["class"]
    aci_rn = ACI_CLASS_MAPPING[contract_type]["rn"]

    if contract_type == "consumer" and provider_match is not None:
        module.fail_json(msg="the 'provider_match' is only configurable for Provided Contracts")

    aci = ACIModule(module)
    aci.construct_url(
        root_class=dict(
            aci_class='fvTenant',
            aci_rn='tn-{0}'.format(tenant),
            filter_target='eq(fvTenant.name, "{0}")'.format(tenant),
            module_object=tenant,
        ),
        subclass_1=dict(
            aci_class='fvAp',
            aci_rn='ap-{0}'.format(ap),
            filter_target='eq(fvAp.name, "{0}")'.format(ap),
            module_object=ap,
        ),
        subclass_2=dict(
            aci_class='fvAEPg',
            aci_rn='epg-{0}'.format(epg),
            filter_target='eq(fvAEPg.name, "{0}")'.format(epg),
            module_object=epg,
        ),
        subclass_3=dict(
            aci_class=aci_class,
            aci_rn='{0}{1}'.format(aci_rn, contract),
            filter_target='eq({0}.tnVzBrCPName, "{1}'.format(aci_class, contract),
            module_object=contract,
        ),
    )

    aci.get_existing()

    if state == 'present':
        aci.payload(
            aci_class=aci_class,
            class_config=dict(
                matchT=provider_match,
                prio=priority,
                tnVzBrCPName=contract,
            ),
        )

        aci.get_diff(aci_class=aci_class)

        aci.post_config()

    elif state == 'absent':
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#
# (c) 2015, Peter Sprygada <psprygada@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.


class ModuleDocFragment(object):

    # Standard files documentation fragment
    DOCUMENTATION = """
options:
  provider:
    description:
      - A dict object containing connection details.
    suboptions:
      host:
        description:
          - Specifies the DNS host name or address for connecting to the remote
            instance of NIOS WAPI over REST
          - Value can also be specified using C(INFOBLOX_HOST) environment
            variable.
        required: true
      username:
        description:
          - Configures the username to use to authenticate the connection to
            the remote instance of NIOS.
          - Value can also be specified using C(INFOBLOX_USERNAME) environment
            variable.
      password:
        description:
          - Specifies the password to use to authenticate the connection to
            the remote instance of NIOS.
          - Value can also be specified using C(INFOBLOX_PASSWORD) environment
            variable.
      ssl_verify:
        description:
          - Boolean value to enable or disable verifying SSL certificates
          - Value can also be specified using C(INFOBLOX_SSL_VERIFY) environment
            variable.
        type: bool
        default: 'no'
      http_request_timeout:
        description:
          - The amount of time before to wait before receiving a response
          - Value can also be specified using C(INFOBLOX_HTTP_REQUEST_TIMEOUT) environment
            variable.
        default: 10
      max_retries:
        description:
          - Configures the number of attempted retries before the connection
            is declared usable
          - Value can also be specified using C(INFOBLOX_MAX_RETRIES) environment
            variable.
        default: 3
      wapi_version:
        description:
          - Specifies the version of WAPI to use
          - Value can also be specified using C(INFOBLOX_WAP_VERSION) environment
            variable.
        default: 1.4
      max_results:
        description:
          - Specifies the maximum number of objects to be returned,
            if set to a negative number the appliance will return an error when the
            number of returned objects would exceed the setting.
          - Value can also be specified using C(INFOBLOX_MAX_RESULTS) environment
            variable.
        default: 1000
notes:
  - "This module must be run locally, which can be achieved by specifying C(connection: local)."
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

#  Copyright (c) 2017 Citrix Systems
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: netscaler_servicegroup
short_description: Manage service group configuration in Netscaler
description:
    - Manage service group configuration in Netscaler.
    - This module is intended to run either on the ansible  control node or a bastion (jumpserver) with access to the actual netscaler instance.

version_added: "2.4"

author: George Nikolopoulos (@giorgos-nikolopoulos)

options:
    servicegroupname:
        description:
            - >-
                Name of the service group. Must begin with an ASCII alphabetic or underscore C(_) character, and must
                contain only ASCII alphanumeric, underscore C(_), hash C(#), period C(.), space C( ), colon C(:), at C(@), equals
                C(=), and hyphen C(-) characters. Can be changed after the name is created.
            - "Minimum length = 1"

    servicetype:
        choices:
            - 'HTTP'
            - 'FTP'
            - 'TCP'
            - 'UDP'
            - 'SSL'
            - 'SSL_BRIDGE'
            - 'SSL_TCP'
            - 'DTLS'
            - 'NNTP'
            - 'RPCSVR'
            - 'DNS'
            - 'ADNS'
            - 'SNMP'
            - 'RTSP'
            - 'DHCPRA'
            - 'ANY'
            - 'SIP_UDP'
            - 'SIP_TCP'
            - 'SIP_SSL'
            - 'DNS_TCP'
            - 'ADNS_TCP'
            - 'MYSQL'
            - 'MSSQL'
            - 'ORACLE'
            - 'RADIUS'
            - 'RADIUSListener'
            - 'RDP'
            - 'DIAMETER'
            - 'SSL_DIAMETER'
            - 'TFTP'
            - 'SMPP'
            - 'PPTP'
            - 'GRE'
            - 'SYSLOGTCP'
            - 'SYSLOGUDP'
            - 'FIX'
            - 'SSL_FIX'
        description:
            - "Protocol used to exchange data with the service."

    cachetype:
        choices:
            - 'TRANSPARENT'
            - 'REVERSE'
            - 'FORWARD'
        description:
            - "Cache type supported by the cache server."

    maxclient:
        description:
            - "Maximum number of simultaneous open connections for the service group."
            - "Minimum value = C(0)"
            - "Maximum value = C(4294967294)"

    maxreq:
        description:
            - "Maximum number of requests that can be sent on a persistent connection to the service group."
            - "Note: Connection requests beyond this value are rejected."
            - "Minimum value = C(0)"
            - "Maximum value = C(65535)"

    cacheable:
        description:
            - "Use the transparent cache redirection virtual server to forward the request to the cache server."
            - "Note: Do not set this parameter if you set the Cache Type."
        type: bool

    cip:
        choices:
            - 'enabled'
            - 'disabled'
        description:
            - "Insert the Client IP header in requests forwarded to the service."

    cipheader:
        description:
            - >-
                Name of the HTTP header whose value must be set to the IP address of the client. Used with the Client
                IP parameter. If client IP insertion is enabled, and the client IP header is not specified, the value
                of Client IP Header parameter or the value set by the set ns config command is used as client's IP
                header name.
            - "Minimum length = 1"

    usip:
        description:
            - >-
                Use client's IP address as the source IP address when initiating connection to the server. With the
                NO setting, which is the default, a mapped IP (MIP) address or subnet IP (SNIP) address is used as
                the source IP address to initiate server side connections.

    pathmonitor:
        description:
            - "Path monitoring for clustering."

    pathmonitorindv:
        description:
            - "Individual Path monitoring decisions."

    useproxyport:
        description:
            - >-
                Use the proxy port as the source port when initiating connections with the server. With the NO
                setting, the client-side connection port is used as the source port for the server-side connection.
            - "Note: This parameter is available only when the Use Source IP C(usip) parameter is set to C(yes)."
        type: bool

    healthmonitor:
        description:
            - "Monitor the health of this service. Available settings function as follows:"
            - "C(yes) - Send probes to check the health of the service."
            - >-
                C(no) - Do not send probes to check the health of the service. With the NO option, the appliance shows
                the service as UP at all times.
        type: bool

    sp:
        description:
            - "Enable surge protection for the service group."
        type: bool

    rtspsessionidremap:
        description:
            - "Enable RTSP session ID mapping for the service group."
        type: bool

    clttimeout:
        description:
            - "Time, in seconds, after which to terminate an idle client connection."
            - "Minimum value = C(0)"
            - "Maximum value = C(31536000)"

    svrtimeout:
        description:
            - "Time, in seconds, after which to terminate an idle server connection."
            - "Minimum value = C(0)"
            - "Maximum value = C(31536000)"

    cka:
        description:
            - "Enable client keep-alive for the service group."
        type: bool

    tcpb:
        description:
            - "Enable TCP buffering for the service group."
        type: bool

    cmp:
        description:
            - "Enable compression for the specified service."
        type: bool

    maxbandwidth:
        description:
            - "Maximum bandwidth, in Kbps, allocated for all the services in the service group."
            - "Minimum value = C(0)"
            - "Maximum value = C(4294967287)"

    monthreshold:
        description:
            - >-
                Minimum sum of weights of the monitors that are bound to this service. Used to determine whether to
                mark a service as UP or DOWN.
            - "Minimum value = C(0)"
            - "Maximum value = C(65535)"

    downstateflush:
        choices:
            - 'enabled'
            - 'disabled'
        description:
            - >-
                Flush all active transactions associated with all the services in the service group whose state
                transitions from UP to DOWN. Do not enable this option for applications that must complete their
                transactions.

    tcpprofilename:
        description:
            - "Name of the TCP profile that contains TCP configuration settings for the service group."
            - "Minimum length = 1"
            - "Maximum length = 127"

    httpprofilename:
        description:
            - "Name of the HTTP profile that contains HTTP configuration settings for the service group."
            - "Minimum length = 1"
            - "Maximum length = 127"

    comment:
        description:
            - "Any information about the service group."

    appflowlog:
        choices:
            - 'enabled'
            - 'disabled'
        description:
            - "Enable logging of AppFlow information for the specified service group."

    netprofile:
        description:
            - "Network profile for the service group."
            - "Minimum length = 1"
            - "Maximum length = 127"

    autoscale:
        choices:
            - 'DISABLED'
            - 'DNS'
            - 'POLICY'
        description:
            - "Auto scale option for a servicegroup."

    memberport:
        description:
            - "member port."

    graceful:
        description:
            - "Wait for all existing connections to the service to terminate before shutting down the service."
        type: bool

    servicemembers:
        description:
            - A list of dictionaries describing each service member of the service group.
        suboptions:
            ip:
                description:
                    - IP address of the service. Must not overlap with an existing server entity defined by name.

            port:
                description:
                    - Server port number.
                    - Range C(1) - C(65535)
                    - "* in CLI is represented as 65535 in NITRO API"
            state:
                choices:
                    - 'enabled'
                    - 'disabled'
                description:
                    - Initial state of the service after binding.
            hashid:
                description:
                    - The hash identifier for the service.
                    - This must be unique for each service.
                    - This parameter is used by hash based load balancing methods.
                    - Minimum value = C(1)

            serverid:
                description:
                    - The identifier for the service.
                    - This is used when the persistency type is set to Custom Server ID.

            servername:
                description:
                    - Name of the server to which to bind the service group.
                    - The server must already be configured as a named server.
                    - Minimum length = 1

            customserverid:
                description:
                    - The identifier for this IP:Port pair.
                    - Used when the persistency type is set to Custom Server ID.

            weight:
                description:
                    - Weight to assign to the servers in the service group.
                    - Specifies the capacity of the servers relative to the other servers in the load balancing configuration.
                    - The higher the weight, the higher the percentage of requests sent to the service.
                    - Minimum value = C(1)
                    - Maximum value = C(100)

    monitorbindings:
        description:
            - A list of monitornames to bind to this service
            - Note that the monitors must have already been setup possibly using the M(netscaler_lb_monitor) module or some other method
        suboptions:
            monitorname:
                description:
                    - The monitor name to bind to this servicegroup.
            weight:
                description:
                    - Weight to assign to the binding between the monitor and servicegroup.

    disabled:
        description:
            - When set to C(yes) the service group state will be set to DISABLED.
            - When set to C(no) the service group state will be set to ENABLED.
            - >-
                Note that due to limitations of the underlying NITRO API a C(disabled) state change alone
                does not cause the module result to report a changed status.
        type: bool
        default: false


extends_documentation_fragment: netscaler
requirements:
    - nitro python sdk
'''

EXAMPLES = '''
# The LB Monitors monitor-1 and monitor-2 must already exist
# Service members defined by C(ip) must not redefine an existing server's ip address.
# Service members defined by C(servername) must already exist.

- name: Setup http service with ip members
  delegate_to: localhost
  netscaler_servicegroup:
    nsip: 172.18.0.2
    nitro_user: nsroot
    nitro_pass: nsroot

    state: present

    servicegroupname: service-group-1
    servicetype: HTTP
    servicemembers:
      - ip: 10.78.78.78
        port: 80
        weight: 50
      - ip: 10.79.79.79
        port: 80
        weight: 40
      - servername: server-1
        port: 80
        weight: 10

    monitorbindings:
      - monitorname: monitor-1
        weight: 50
      - monitorname: monitor-2
        weight: 50

'''

RETURN = '''
loglines:
    description: list of logged messages by the module
    returned: always
    type: list
    sample: ['message 1', 'message 2']

msg:
    description: Message detailing the failure reason
    returned: failure
    type: str
    sample: "Action does not exist"

diff:
    description: List of differences between the actual configured object and the configuration specified in the module
    returned: failure
    type: dict
    sample: { 'clttimeout': 'difference. ours: (float) 10.0 other: (float) 20.0' }
'''

from ansible.module_utils.basic import AnsibleModule
import copy

from ansible.module_utils.network.netscaler.netscaler import ConfigProxy, get_nitro_client, netscaler_common_arguments, log, \
    loglines, get_immutables_intersection
try:
    from nssrc.com.citrix.netscaler.nitro.resource.config.basic.servicegroup import servicegroup
    from nssrc.com.citrix.netscaler.nitro.resource.config.basic.servicegroup_servicegroupmember_binding import servicegroup_servicegroupmember_binding
    from nssrc.com.citrix.netscaler.nitro.exception.nitro_exception import nitro_exception

    from nssrc.com.citrix.netscaler.nitro.resource.config.basic.servicegroup_lbmonitor_binding import servicegroup_lbmonitor_binding
    from nssrc.com.citrix.netscaler.nitro.resource.config.lb.lbmonitor_servicegroup_binding import lbmonitor_servicegroup_binding
    PYTHON_SDK_IMPORTED = True
except ImportError as e:
    PYTHON_SDK_IMPORTED = False


def servicegroup_exists(client, module):
    log('Checking if service group exists')
    count = servicegroup.count_filtered(client, 'servicegroupname:%s' % module.params['servicegroupname'])
    log('count is %s' % count)
    if count > 0:
        return True
    else:
        return False


def servicegroup_identical(client, module, servicegroup_proxy):
    log('Checking if service group is identical')
    servicegroups = servicegroup.get_filtered(client, 'servicegroupname:%s' % module.params['servicegroupname'])
    if servicegroup_proxy.has_equal_attributes(servicegroups[0]):
        return True
    else:
        return False


def get_configured_service_members(client, module):
    log('get_configured_service_members')
    readwrite_attrs = [
        'servicegroupname',
        'ip',
        'port',
        'state',
        'hashid',
        'serverid',
        'servername',
        'customserverid',
        'weight'
    ]
    readonly_attrs = [
        'delay',
        'statechangetimesec',
        'svrstate',
        'tickssincelaststatechange',
        'graceful',
    ]

    members = []
    if module.params['servicemembers'] is None:
        return members

    for config in module.params['servicemembers']:
        # Make a copy to update
        config = copy.deepcopy(config)
        config['servicegroupname'] = module.params['servicegroupname']
        member_proxy = ConfigProxy(
            actual=servicegroup_servicegroupmember_binding(),
            client=client,
            attribute_values_dict=config,
            readwrite_attrs=readwrite_attrs,
            readonly_attrs=readonly_attrs
        )
        members.append(member_proxy)
    return members


def get_actual_service_members(client, module):
    try:
        # count() raises nitro exception instead of returning 0
        count = servicegroup_servicegroupmember_binding.count(client, module.params['servicegroupname'])
        if count > 0:
            servicegroup_members = servicegroup_servicegroupmember_binding.get(client, module.params['servicegroupname'])
        else:
            servicegroup_members = []
    except nitro_exception as e:
        if e.errorcode == 258:
            servicegroup_members = []
        else:
            raise
    return servicegroup_members


def servicemembers_identical(client, module):
    log('servicemembers_identical')

    servicegroup_members = get_actual_service_members(client, module)
    log('servicemembers %s' % servicegroup_members)
    module_servicegroups = get_configured_service_members(client, module)
    log('Number of service group members %s' % len(servicegroup_members))
    if len(servicegroup_members) != len(module_servicegroups):
        return False

    # Fallthrough to member evaluation
    identical_count = 0
    for actual_member in servicegroup_members:
        for member in module_servicegroups:
            if member.has_equal_attributes(actual_member):
                identical_count += 1
                break
    if identical_count != len(servicegroup_members):
        return False

    # Fallthrough to success
    return True


def sync_service_members(client, module):
    log('sync_service_members')
    configured_service_members = get_configured_service_members(client, module)
    actual_service_members = get_actual_service_members(client, module)
    skip_add = []
    skip_delete = []

    # Find positions of identical service members
    for (configured_index, configured_service) in enumerate(configured_service_members):
        for (actual_index, actual_service) in enumerate(actual_service_members):
            if configured_service.has_equal_attributes(actual_service):
                skip_add.append(configured_index)
                skip_delete.append(actual_index)

    # Delete actual that are not identical to any configured
    for (actual_index, actual_service) in enumerate(actual_service_members):
        # Skip identical
        if actual_index in skip_delete:
            log('Skipping actual delete at index %s' % actual_index)
            continue

        # Fallthrouth to deletion
        if all([
            hasattr(actual_service, 'ip'),
            actual_service.ip is not None,
            hasattr(actual_service, 'servername'),
            actual_service.servername is not None,
        ]):
            actual_service.ip = None

        actual_service.servicegroupname = module.params['servicegroupname']
        servicegroup_servicegroupmember_binding.delete(client, actual_service)

    # Add configured that are not already present in actual
    for (configured_index, configured_service) in enumerate(configured_service_members):

        # Skip identical
        if configured_index in skip_add:
            log('Skipping configured add at index %s' % configured_index)
            continue

        # Fallthrough to addition
        configured_service.add()


def monitor_binding_equal(configured, actual):
    if any([configured.monitorname != actual.monitor_name,
            configured.servicegroupname != actual.servicegroupname,
            configured.weight != float(actual.weight)]):
        return False
    return True


def get_configured_monitor_bindings(client, module):
    log('Entering get_configured_monitor_bindings')
    bindings = {}
    if 'monitorbindings' in module.params and module.params['monitorbindings'] is not None:
        for binding in module.params['monitorbindings']:
            readwrite_attrs = [
                'monitorname',
                'servicegroupname',
                'weight',
            ]
            readonly_attrs = []
            attribute_values_dict = copy.deepcopy(binding)
            attribute_values_dict['servicegroupname'] = module.params['servicegroupname']
            binding_proxy = ConfigProxy(
                actual=lbmonitor_servicegroup_binding(),
                client=client,
                attribute_values_dict=attribute_values_dict,
                readwrite_attrs=readwrite_attrs,
                readonly_attrs=readonly_attrs,
            )
            key = attribute_values_dict['monitorname']
            bindings[key] = binding_proxy
    return bindings


def get_actual_monitor_bindings(client, module):
    log('Entering get_actual_monitor_bindings')
    bindings = {}
    try:
        # count() raises nitro exception instead of returning 0
        count = servicegroup_lbmonitor_binding.count(client, module.params['servicegroupname'])
    except nitro_exception as e:
        if e.errorcode == 258:
            return bindings
        else:
            raise

    if count == 0:
        return bindings

    # Fallthrough to rest of execution
    for binding in servicegroup_lbmonitor_binding.get(client, module.params['servicegroupname']):
        log('Gettign actual monitor with name %s' % binding.monitor_name)
        key = binding.monitor_name
        bindings[key] = binding

    return bindings


def monitor_bindings_identical(client, module):
    log('Entering monitor_bindings_identical')
    configured_bindings = get_configured_monitor_bindings(client, module)
    actual_bindings = get_actual_monitor_bindings(client, module)

    configured_key_set = set(configured_bindings.keys())
    actual_key_set = set(actual_bindings.keys())
    symmetrical_diff = configured_key_set ^ actual_key_set
    for default_monitor in ('tcp-default', 'ping-default'):
        if default_monitor in symmetrical_diff:
            log('Excluding %s monitor from key comparison' % default_monitor)
            symmetrical_diff.remove(default_monitor)
    if len(symmetrical_diff) > 0:
        return False

    # Compare key to key
    for key in configured_key_set:
        configured_proxy = configured_bindings[key]

        # Follow nscli convention for missing weight value
        if not hasattr(configured_proxy, 'weight'):
            configured_proxy.weight = 1
        log('configured_proxy %s' % [configured_proxy.monitorname, configured_proxy.servicegroupname, configured_proxy.weight])
        log('actual_bindings %s' % [actual_bindings[key].monitor_name, actual_bindings[key].servicegroupname, actual_bindings[key].weight])
        if not monitor_binding_equal(configured_proxy, actual_bindings[key]):
            return False

    # Fallthrought to success
    return True


def sync_monitor_bindings(client, module):
    log('Entering sync_monitor_bindings')

    actual_bindings = get_actual_monitor_bindings(client, module)

    # Exclude default monitors from deletion
    for monitorname in ('tcp-default', 'ping-default'):
        if monitorname in actual_bindings:
            del actual_bindings[monitorname]

    configured_bindings = get_configured_monitor_bindings(client, module)

    to_remove = list(set(actual_bindings.keys()) - set(configured_bindings.keys()))
    to_add = list(set(configured_bindings.keys()) - set(actual_bindings.keys()))
    to_modify = list(set(configured_bindings.keys()) & set(actual_bindings.keys()))

    # Delete existing and modifiable bindings
    for key in to_remove + to_modify:
        binding = actual_bindings[key]
        b = lbmonitor_servicegroup_binding()
        b.monitorname = binding.monitor_name
        b.servicegroupname = module.params['servicegroupname']
        # Cannot remove default monitor bindings
        if b.monitorname in ('tcp-default', 'ping-default'):
            continue
        lbmonitor_servicegroup_binding.delete(client, b)

    # Add new and modified bindings
    for key in to_add + to_modify:
        binding = configured_bindings[key]
        log('Adding %s' % binding.monitorname)
        binding.add()


def diff(client, module, servicegroup_proxy):
    servicegroup_list = servicegroup.get_filtered(client, 'servicegroupname:%s' % module.params['servicegroupname'])
    diff_object = servicegroup_proxy.diff_object(servicegroup_list[0])
    return diff_object


def do_state_change(client, module, servicegroup_proxy):
    if module.params['disabled']:
        log('Disabling service')
        result = servicegroup.disable(client, servicegroup_proxy.actual)
    else:
        log('Enabling service')
        result = servicegroup.enable(client, servicegroup_proxy.actual)
    return result


def main():

    module_specific_arguments = dict(
        servicegroupname=dict(type='str'),
        servicetype=dict(
            type='str',
            choices=[
                'HTTP',
                'FTP',
                'TCP',
                'UDP',
                'SSL',
                'SSL_BRIDGE',
                'SSL_TCP',
                'DTLS',
                'NNTP',
                'RPCSVR',
                'DNS',
                'ADNS',
                'SNMP',
                'RTSP',
                'DHCPRA',
                'ANY',
                'SIP_UDP',
                'SIP_TCP',
                'SIP_SSL',
                'DNS_TCP',
                'ADNS_TCP',
                'MYSQL',
                'MSSQL',
                'ORACLE',
                'RADIUS',
                'RADIUSListener',
                'RDP',
                'DIAMETER',
                'SSL_DIAMETER',
                'TFTP',
                'SMPP',
                'PPTP',
                'GRE',
                'SYSLOGTCP',
                'SYSLOGUDP',
                'FIX',
                'SSL_FIX',
            ]
        ),
        cachetype=dict(
            type='str',
            choices=[
                'TRANSPARENT',
                'REVERSE',
                'FORWARD',
            ]
        ),
        maxclient=dict(type='float'),
        maxreq=dict(type='float'),
        cacheable=dict(type='bool'),
        cip=dict(
            type='str',
            choices=[
                'enabled',
                'disabled',
            ]
        ),
        cipheader=dict(type='str'),
        usip=dict(type='bool'),
        pathmonitor=dict(type='bool'),
        pathmonitorindv=dict(type='bool'),
        useproxyport=dict(type='bool'),
        healthmonitor=dict(type='bool'),
        sp=dict(type='bool'),
        rtspsessionidremap=dict(type='bool'),
        clttimeout=dict(type='float'),
        svrtimeout=dict(type='float'),
        cka=dict(type='bool'),
        tcpb=dict(type='bool'),
        cmp=dict(type='bool'),
        maxbandwidth=dict(type='float'),
        monthreshold=dict(type='float'),
        downstateflush=dict(
            type='str',
            choices=[
                'enabled',
                'disabled',
            ]
        ),
        tcpprofilename=dict(type='str'),
        httpprofilename=dict(type='str'),
        comment=dict(type='str'),
        appflowlog=dict(
            type='str',
            choices=[
                'enabled',
                'disabled',
            ]
        ),
        netprofile=dict(type='str'),
        autoscale=dict(
            type='str',
            choices=[
                'DISABLED',
                'DNS',
                'POLICY',
            ]
        ),
        memberport=dict(type='int'),
        graceful=dict(type='bool'),
    )

    hand_inserted_arguments = dict(
        servicemembers=dict(type='list'),
        monitorbindings=dict(type='list'),
        disabled=dict(
            type='bool',
            default=False,
        ),
    )

    argument_spec = dict()

    argument_spec.update(netscaler_common_arguments)
    argument_spec.update(module_specific_arguments)
    argument_spec.update(hand_inserted_arguments)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )
    module_result = dict(
        changed=False,
        failed=False,
        loglines=loglines,
    )

    # Fail the module if imports failed
    if not PYTHON_SDK_IMPORTED:
        module.fail_json(msg='Could not load nitro python sdk')

    # Fallthrough to rest of execution
    client = get_nitro_client(module)

    try:
        client.login()
    except nitro_exception as e:
        msg = "nitro exception during login. errorcode=%s, message=%s" % (str(e.errorcode), e.message)
        module.fail_json(msg=msg)
    except Exception as e:
        if str(type(e)) == "<class 'requests.exceptions.ConnectionError'>":
            module.fail_json(msg='Connection error %s' % str(e))
        elif str(type(e)) == "<class 'requests.exceptions.SSLError'>":
            module.fail_json(msg='SSL Error %s' % str(e))
        else:
            module.fail_json(msg='Unexpected error during login %s' % str(e))

    # Instantiate service group configuration object
    readwrite_attrs = [
        'servicegroupname',
        'servicetype',
        'cachetype',
        'maxclient',
        'maxreq',
        'cacheable',
        'cip',
        'cipheader',
        'usip',
        'pathmonitor',
        'pathmonitorindv',
        'useproxyport',
        'healthmonitor',
        'sp',
        'rtspsessionidremap',
        'clttimeout',
        'svrtimeout',
        'cka',
        'tcpb',
        'cmp',
        'maxbandwidth',
        'monthreshold',
        'downstateflush',
        'tcpprofilename',
        'httpprofilename',
        'comment',
        'appflowlog',
        'netprofile',
        'autoscale',
        'memberport',
        'graceful',
    ]

    readonly_attrs = [
        'numofconnections',
        'serviceconftype',
        'value',
        'svrstate',
        'ip',
        'monstatcode',
        'monstatparam1',
        'monstatparam2',
        'monstatparam3',
        'statechangetimemsec',
        'stateupdatereason',
        'clmonowner',
        'clmonview',
        'groupcount',
        'riseapbrstatsmsgcode2',
        'serviceipstr',
        'servicegroupeffectivestate'
    ]

    immutable_attrs = [
        'servicegroupname',
        'servicetype',
        'cachetype',
        'td',
        'cipheader',
        'state',
        'autoscale',
        'memberport',
        'servername',
        'port',
        'serverid',
        'monitor_name_svc',
        'dup_weight',
        'riseapbrstatsmsgcode',
        'delay',
        'graceful',
        'includemembers',
        'newname',
    ]

    transforms = {
        'pathmonitorindv': ['bool_yes_no'],
        'cacheable': ['bool_yes_no'],
        'cka': ['bool_yes_no'],
        'pathmonitor': ['bool_yes_no'],
        'tcpb': ['bool_yes_no'],
        'sp': ['bool_on_off'],
        'usip': ['bool_yes_no'],
        'healthmonitor': ['bool_yes_no'],
        'useproxyport': ['bool_yes_no'],
        'rtspsessionidremap': ['bool_on_off'],
        'graceful': ['bool_yes_no'],
        'cmp': ['bool_yes_no'],
        'cip': [lambda v: v.upper()],
        'downstateflush': [lambda v: v.upper()],
        'appflowlog': [lambda v: v.upper()],
    }

    # Instantiate config proxy
    servicegroup_proxy = ConfigProxy(
        actual=servicegroup(),
        client=client,
        attribute_values_dict=module.params,
        readwrite_attrs=readwrite_attrs,
        readonly_attrs=readonly_attrs,
        immutable_attrs=immutable_attrs,
        transforms=transforms,
    )

    try:
        if module.params['state'] == 'present':
            log('Applying actions for state present')
            if not servicegroup_exists(client, module):
                if not module.check_mode:
                    log('Adding service group')
                    servicegroup_proxy.add()
                    if module.params['save_config']:
                        client.save_config()
                module_result['changed'] = True
            elif not servicegroup_identical(client, module, servicegroup_proxy):

                # Check if we try to change value of immutable attributes
                diff_dict = diff(client, module, servicegroup_proxy)
                immutables_changed = get_immutables_intersection(servicegroup_proxy, diff_dict.keys())
                if immutables_changed != []:
                    msg = 'Cannot update immutable attributes %s. Must delete and recreate entity.' % (immutables_changed,)
                    module.fail_json(msg=msg, diff=diff_dict, **module_result)

                if not module.check_mode:
                    servicegroup_proxy.update()
                    if module.params['save_config']:
                        client.save_config()
                module_result['changed'] = True
            else:
                module_result['changed'] = False

            # Check bindings
            if not monitor_bindings_identical(client, module):
                if not module.check_mode:
                    sync_monitor_bindings(client, module)
                    if module.params['save_config']:
                        client.save_config()
                module_result['changed'] = True

            if not servicemembers_identical(client, module):
                if not module.check_mode:
                    sync_service_members(client, module)
                    if module.params['save_config']:
                        client.save_config()
                module_result['changed'] = True

            if not module.check_mode:
                res = do_state_change(client, module, servicegroup_proxy)
                if res.errorcode != 0:
                    msg = 'Error when setting disabled state. errorcode: %s message: %s' % (res.errorcode, res.message)
                    module.fail_json(msg=msg, **module_result)

            # Sanity check for state
            if not module.check_mode:
                log('Sanity checks for state present')
                if not servicegroup_exists(client, module):
                    module.fail_json(msg='Service group is not present', **module_result)
                if not servicegroup_identical(client, module, servicegroup_proxy):
                    module.fail_json(
                        msg='Service group is not identical to configuration',
                        diff=diff(client, module, servicegroup_proxy),
                        **module_result
                    )
                if not servicemembers_identical(client, module):
                    module.fail_json(msg='Service group members differ from configuration', **module_result)
                if not monitor_bindings_identical(client, module):
                    module.fail_json(msg='Monitor bindings are not identical', **module_result)

        elif module.params['state'] == 'absent':
            log('Applying actions for state absent')
            if servicegroup_exists(client, module):
                if not module.check_mode:
                    servicegroup_proxy.delete()
                    if module.params['save_config']:
                        client.save_config()
                module_result['changed'] = True
            else:
                module_result['changed'] = False

            # Sanity check for state
            if not module.check_mode:
                log('Sanity checks for state absent')
                if servicegroup_exists(client, module):
                    module.fail_json(msg='Service group is present', **module_result)

    except nitro_exception as e:
        msg = "nitro exception errorcode=" + str(e.errorcode) + ",message=" + e.message
        module.fail_json(msg=msg, **module_result)

    client.logout()
    module.exit_json(**module_result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2016-2017 Hewlett Packard Enterprise Development LP
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = '''
---
module: oneview_datacenter_facts
short_description: Retrieve facts about the OneView Data Centers
description:
    - Retrieve facts about the OneView Data Centers.
version_added: "2.5"
requirements:
    - "hpOneView >= 2.0.1"
author:
    - Alex Monteiro (@aalexmonteiro)
    - Madhav Bharadwaj (@madhav-bharadwaj)
    - Priyanka Sood (@soodpr)
    - Ricardo Galeno (@ricardogpsf)
options:
    name:
      description:
        - Data Center name.
    options:
      description:
        - "Retrieve additional facts. Options available: 'visualContent'."

extends_documentation_fragment:
    - oneview
    - oneview.factsparams
'''

EXAMPLES = '''
- name: Gather facts about all Data Centers
  oneview_datacenter_facts:
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
  delegate_to: localhost
- debug: var=datacenters

- name: Gather paginated, filtered and sorted facts about Data Centers
  oneview_datacenter_facts:
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
    params:
      start: 0
      count: 3
      sort: 'name:descending'
      filter: 'state=Unmanaged'
- debug: var=datacenters

- name: Gather facts about a Data Center by name
  oneview_datacenter_facts:
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
    name: "My Data Center"
  delegate_to: localhost
- debug: var=datacenters

- name: Gather facts about the Data Center Visual Content
  oneview_datacenter_facts:
    hostname: 172.16.101.48
    username: administrator
    password: my_password
    api_version: 500
    name: "My Data Center"
    options:
      - visualContent
  delegate_to: localhost
- debug: var=datacenters
- debug: var=datacenter_visual_content
'''

RETURN = '''
datacenters:
    description: Has all the OneView facts about the Data Centers.
    returned: Always, but can be null.
    type: dict

datacenter_visual_content:
    description: Has facts about the Data Center Visual Content.
    returned: When requested, but can be null.
    type: dict
'''

from ansible.module_utils.oneview import OneViewModuleBase


class DatacenterFactsModule(OneViewModuleBase):
    argument_spec = dict(
        name=dict(type='str'),
        options=dict(type='list'),
        params=dict(type='dict')
    )

    def __init__(self):
        super(DatacenterFactsModule, self).__init__(additional_arg_spec=self.argument_spec)

    def execute_module(self):

        client = self.oneview_client.datacenters
        ansible_facts = {}

        if self.module.params.get('name'):
            datacenters = client.get_by('name', self.module.params['name'])

            if self.options and 'visualContent' in self.options:
                if datacenters:
                    ansible_facts['datacenter_visual_content'] = client.get_visual_content(datacenters[0]['uri'])
                else:
                    ansible_facts['datacenter_visual_content'] = None

            ansible_facts['datacenters'] = datacenters
        else:
            ansible_facts['datacenters'] = client.get_all(**self.facts_params)

        return dict(changed=False,
                    ansible_facts=ansible_facts)


def main():
    DatacenterFactsModule().run()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2016 Red Hat, Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#


class ModuleDocFragment(object):

    # facts standard oVirt documentation fragment
    DOCUMENTATION = '''
options:
    fetch_nested:
        description:
            - "If I(True) the module will fetch additional data from the API."
            - "It will fetch IDs of the VMs disks, snapshots, etc. User can configure to fetch other
               attributes of the nested entities by specifying C(nested_attributes)."
        version_added: "2.3"
    nested_attributes:
        description:
            - "Specifies list of the attributes which should be fetched from the API."
            - "This parameter apply only when C(fetch_nested) is I(true)."
        version_added: "2.3"
    auth:
        required: True
        description:
            - "Dictionary with values needed to create HTTP/HTTPS connection to oVirt:"
            - C(username)[I(required)] - The name of the user, something like I(admin@internal).
              Default value is set by I(OVIRT_USERNAME) environment variable.
            - "C(password)[I(required)] - The password of the user. Default value is set by I(OVIRT_PASSWORD) environment variable."
            - "C(url)- A string containing the API URL of the server, usually
            something like `I(https://server.example.com/ovirt-engine/api)`. Default value is set by I(OVIRT_URL) environment variable.
            Either C(url) or C(hostname) is required."
            - "C(hostname) - A string containing the hostname of the server, usually
            something like `I(server.example.com)`. Default value is set by I(OVIRT_HOSTNAME) environment variable.
            Either C(url) or C(hostname) is required."
            - "C(token) - Token to be used instead of login with username/password. Default value is set by I(OVIRT_TOKEN) environment variable."
            - "C(insecure) - A boolean flag that indicates if the server TLS
            certificate and host name should be checked."
            - "C(ca_file) - A PEM file containing the trusted CA certificates. The
            certificate presented by the server will be verified using these CA
            certificates. If `C(ca_file)` parameter is not set, system wide
            CA certificate store is used. Default value is set by I(OVIRT_CAFILE) environment variable."
            - "C(kerberos) - A boolean flag indicating if Kerberos authentication
            should be used instead of the default basic authentication."
            - "C(headers) - Dictionary of HTTP headers to be added to each API call."
requirements:
  - python >= 2.7
  - ovirt-engine-sdk-python >= 4.2.4
notes:
  - "In order to use this module you have to install oVirt Python SDK.
     To ensure it's installed with correct version you can create the following task:
     pip: name=ovirt-engine-sdk-python version=4.2.4"
'''



#!/usr/bin/env python

import os


def main():
    # required by external automated processes and should not be moved, renamed or converted to a symbolic link
    path = 'examples/scripts/ConfigureRemotingForAnsible.ps1'
    directory = path

    while True:
        directory = os.path.dirname(directory)

        if not directory:
            break

        if not os.path.isdir(directory):
            print('%s: must be a directory' % directory)

        if os.path.islink(directory):
            print('%s: cannot be a symbolic link' % directory)

    if not os.path.isfile(path):
        print('%s: must be a file' % path)

    if os.path.islink(path):
        print('%s: cannot be a symbolic link' % path)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: ucs_lan_connectivity
short_description: Configures LAN Connectivity Policies on Cisco UCS Manager
description:
- Configures LAN Connectivity Policies on Cisco UCS Manager.
- Examples can be used with the UCS Platform Emulator U(https://communities.cisco.com/ucspe).
extends_documentation_fragment: ucs
options:
  state:
    description:
    - If C(present), will verify LAN Connectivity Policies are present and will create if needed.
    - If C(absent), will verify LAN Connectivity Policies are absent and will delete if needed.
    choices: [present, absent]
    default: present
  name:
    description:
    - The name of the LAN Connectivity Policy.
    - This name can be between 1 and 16 alphanumeric characters.
    - "You cannot use spaces or any special characters other than - (hyphen), \"_\" (underscore), : (colon), and . (period)."
    - You cannot change this name after the policy is created.
    required: yes
  description:
    description:
    - A description of the LAN Connectivity Policy.
    - Cisco recommends including information about where and when to use the policy.
    - Enter up to 256 characters.
    - "You can use any characters or spaces except the following:"
    - "` (accent mark), \ (backslash), ^ (carat), \" (double quote), = (equal sign), > (greater than), < (less than), or ' (single quote)."
    aliases: [ descr ]
  vnic_list:
    description:
    - List of vNICs used by the LAN Connectivity Policy.
    - vNICs used by the LAN Connectivity Policy must be created from a vNIC template.
    - "Each list element has the following suboptions:"
    - "= name"
    - "  The name of the vNIC (required)."
    - "= vnic_template"
    - "  The name of the vNIC template (required)."
    - "- adapter_policy"
    - "  The name of the Ethernet adapter policy."
    - "  A user defined policy can be used, or one of the system defined policies."
    - "- order"
    - "  String specifying the vNIC assignment order (e.g., '1', '2')."
    - "  [Default: unspecified]"
  org_dn:
    description:
    - Org dn (distinguished name)
    default: org-root
requirements:
- ucsmsdk
author:
- David Soper (@dsoper2)
- CiscoUcs (@CiscoUcs)
version_added: '2.5'
'''

EXAMPLES = r'''
- name: Configure LAN Connectivity Policy
  ucs_lan_connectivity:
    hostname: 172.16.143.150
    username: admin
    password: password
    name: Cntr-LAN-Boot
    vnic_list:
    - name: Fabric-A
      vnic_template: vNIC-Template-A
      adapter_policy: Linux
    - name: Fabric-B
      vnic_template: vNIC-Template-B
      adapter_policy: Linux

- name: Remove LAN Connectivity Policy
  ucs_lan_connectivity:
    hostname: 172.16.143.150
    username: admin
    password: password
    name: Cntr-LAN-Boot
    state: absent
'''

RETURN = r'''
#
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.remote_management.ucs import UCSModule, ucs_argument_spec


def main():
    argument_spec = ucs_argument_spec
    argument_spec.update(
        org_dn=dict(type='str', default='org-root'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr'], default=''),
        vnic_list=dict(type='list'),
        state=dict(type='str', default='present', choices=['present', 'absent']),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )
    ucs = UCSModule(module)

    err = False

    # UCSModule creation above verifies ucsmsdk is present and exits on failure.  Additional imports are done below.
    from ucsmsdk.mometa.vnic.VnicLanConnPolicy import VnicLanConnPolicy
    from ucsmsdk.mometa.vnic.VnicEther import VnicEther
    from ucsmsdk.mometa.vnic.VnicEtherIf import VnicEtherIf

    changed = False
    try:
        mo_exists = False
        props_match = False
        # dn is <org_dn>/lan-conn-pol-<name>
        dn = module.params['org_dn'] + '/lan-conn-pol-' + module.params['name']

        mo = ucs.login_handle.query_dn(dn)
        if mo:
            mo_exists = True

        if module.params['state'] == 'absent':
            # mo must exist but all properties do not have to match
            if mo_exists:
                if not module.check_mode:
                    ucs.login_handle.remove_mo(mo)
                    ucs.login_handle.commit()
                changed = True
        else:
            # set default params.  Done here to set values for lists which can't be done in the argument_spec
            if module.params.get('vnic_list'):
                for vnic in module.params['vnic_list']:
                    if not vnic.get('adapter_policy'):
                        vnic['adapter_policy'] = ''
                    if not vnic.get('order'):
                        vnic['order'] = 'unspecified'
            if mo_exists:
                # check top-level mo props
                kwargs = dict(descr=module.params['description'])
                if (mo.check_prop_match(**kwargs)):
                    # top-level props match, check next level mo/props
                    if not module.params.get('vnic_list'):
                        props_match = True
                    else:
                        # check vnicEther props
                        for vnic in module.params['vnic_list']:
                            child_dn = dn + '/ether-' + vnic['name']
                            mo_1 = ucs.login_handle.query_dn(child_dn)
                            if mo_1:
                                kwargs = dict(adaptor_profile_name=vnic['adapter_policy'])
                                kwargs['order'] = vnic['order']
                                kwargs['nw_templ_name'] = vnic['vnic_template']
                                if (mo_1.check_prop_match(**kwargs)):
                                    props_match = True

            if not props_match:
                if not module.check_mode:
                    # create if mo does not already exist
                    mo = VnicLanConnPolicy(
                        parent_mo_or_dn=module.params['org_dn'],
                        name=module.params['name'],
                        descr=module.params['description'],
                    )

                    if module.params.get('vnic_list'):
                        for vnic in module.params['vnic_list']:
                            mo_1 = VnicEther(
                                addr='derived',
                                parent_mo_or_dn=mo,
                                name=vnic['name'],
                                adaptor_profile_name=vnic['adapter_policy'],
                                nw_templ_name=vnic['vnic_template'],
                                order=vnic['order'],
                            )

                    ucs.login_handle.add_mo(mo, True)
                    ucs.login_handle.commit()
                changed = True

    except Exception as e:
        err = True
        ucs.result['msg'] = "setup error: %s " % str(e)

    ucs.result['changed'] = changed
    if err:
        module.fail_json(**ucs.result)
    module.exit_json(**ucs.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# this is a windows documentation stub.  actual code lives in the .ps1
# file of the same name

# Copyright: (c) 2018, Ripon Banik (@riponbanik)
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: win_hostname
version_added: "2.6"
short_description: Manages local Windows computer name.
description:
- Manages local Windows computer name.
- A reboot is required for the computer name to take effect.
options:
  name:
    description:
    - The hostname to set for the computer.
    required: true
author:
- Ripon Banik (@riponbanik)
'''

EXAMPLES = r'''
- name: Change the hostname to new_hostname
  win_hostname:
    name: new_hostname
'''

RETURN = r'''
old_name:
  description: The original hostname that was set before it was changed
  returned: always
  type: str
  sample: old_hostname
reboot_required:
  description: Whether a reboot is required to complete the hostname change
  returned: always
  type: bool
  sample: true
'''



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017 F5 Networks Inc.
# GNU General Public License v3.0 (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['stableinterface'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: bigip_iapp_service
short_description: Manages TCL iApp services on a BIG-IP
description:
  - Manages TCL iApp services on a BIG-IP.
  - If you are looking for the API that is communicated with on the BIG-IP,
    the one the is used is C(/mgmt/tm/sys/application/service/). There are a
    couple of APIs in a BIG-IP that might seem like they are relevant to iApp
    Services, but the API mentioned here is the one that is used.
version_added: 2.4
options:
  name:
    description:
      - The name of the iApp service that you want to deploy.
    required: True
  template:
    description:
      - The iApp template from which to instantiate a new service. This
        template must exist on your BIG-IP before you can successfully
        create a service. This parameter is required if the C(state)
        parameter is C(present).
  parameters:
    description:
      - A hash of all the required template variables for the iApp template.
        If your parameters are stored in a file (the more common scenario)
        it is recommended you use either the `file` or `template` lookups
        to supply the expected parameters.
      - These parameters typically consist of the C(lists), C(tables), and
        C(variables) fields.
  force:
    description:
      - Forces the updating of an iApp service even if the parameters to the
        service have not changed. This option is of particular importance if
        the iApp template that underlies the service has been updated in-place.
        This option is equivalent to re-configuring the iApp if that template
        has changed.
    default: no
    type: bool
  state:
    description:
      - When C(present), ensures that the iApp service is created and running.
        When C(absent), ensures that the iApp service has been removed.
    default: present
    choices:
      - present
      - absent
  partition:
    description:
      - Device partition to manage resources on.
    default: Common
    version_added: 2.5
  strict_updates:
    description:
      - Indicates whether the application service is tied to the template,
        so when the template is updated, the application service changes to
        reflect the updates.
      - When C(yes), disallows any updates to the resources that the iApp
        service has created, if they are not updated directly through the
        iApp.
      - When C(no), allows updates outside of the iApp.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Server payload that you provide in
        the C(parameters) field.
    default: yes
    type: bool
    version_added: 2.5
  traffic_group:
    description:
      - The traffic group for the iApp service. When creating a new service, if
        this value is not specified, the default of C(/Common/traffic-group-1)
        will be used.
      - If this option is specified in the Ansible task, it will take precedence
        over any similar setting in the iApp Server payload that you provide in
        the C(parameters) field.
    version_added: 2.5
extends_documentation_fragment: f5
author:
  - Tim Rupp (@caphrim007)
'''

EXAMPLES = r'''
- name: Create HTTP iApp service from iApp template
  bigip_iapp_service:
    name: foo-service
    template: f5.http
    parameters: "{{ lookup('file', 'f5.http.parameters.json') }}"
    password: secret
    server: lb.mydomain.com
    state: present
    user: admin
  delegate_to: localhost

- name: Upgrade foo-service to v1.2.0rc4 of the f5.http template
  bigip_iapp_service:
    name: foo-service
    template: f5.http.v1.2.0rc4
    password: secret
    server: lb.mydomain.com
    state: present
    user: admin
  delegate_to: localhost

- name: Configure a service using parameters in YAML
  bigip_iapp_service:
    name: tests
    template: web_frontends
    password: admin
    server: "{{ inventory_hostname }}"
    server_port: "{{ bigip_port }}"
    validate_certs: "{{ validate_certs }}"
    state: present
    user: admin
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
  delegate_to: localhost

- name: Re-configure a service whose underlying iApp was updated in place
  bigip_iapp_service:
    name: tests
    template: web_frontends
    password: admin
    force: yes
    server: "{{ inventory_hostname }}"
    server_port: "{{ bigip_port }}"
    validate_certs: "{{ validate_certs }}"
    state: present
    user: admin
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
  delegate_to: localhost

- name: Try to remove the iApp template before the associated Service is removed
  bigip_iapp_template:
    name: web_frontends
    state: absent
  register: result
  failed_when:
    - result is not success
    - "'referenced by one or more applications' not in result.msg"

- name: Configure a service using more complicated parameters
  bigip_iapp_service:
    name: tests
    template: web_frontends
    password: admin
    server: "{{ inventory_hostname }}"
    server_port: "{{ bigip_port }}"
    validate_certs: "{{ validate_certs }}"
    state: present
    user: admin
    parameters:
      variables:
        - name: var__vs_address
          value: 1.1.1.1
        - name: pm__apache_servers_for_http
          value: 2.2.2.1:80
        - name: pm__apache_servers_for_https
          value: 2.2.2.2:80
      lists:
        - name: irules__irules
          value:
            - foo
            - bar
      tables:
        - name: basic__snatpool_members
        - name: net__snatpool_members
        - name: optimizations__hosts
        - name: pool__hosts
          columnNames:
            - name
          rows:
            - row:
                - internal.company.bar
        - name: pool__members
          columnNames:
            - addr
            - port
            - connection_limit
          rows:
            - row:
                - "none"
                - 80
                - 0
        - name: server_pools__servers
  delegate_to: localhost
'''

RETURN = r'''
# only common fields returned
'''

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.basic import env_fallback
from ansible.module_utils.six import iteritems

try:
    from library.module_utils.network.f5.bigip import HAS_F5SDK
    from library.module_utils.network.f5.bigip import F5Client
    from library.module_utils.network.f5.common import F5ModuleError
    from library.module_utils.network.f5.common import AnsibleF5Parameters
    from library.module_utils.network.f5.common import cleanup_tokens
    from library.module_utils.network.f5.common import fq_name
    from library.module_utils.network.f5.common import f5_argument_spec
    try:
        from library.module_utils.network.f5.common import iControlUnexpectedHTTPError
    except ImportError:
        HAS_F5SDK = False
except ImportError:
    from ansible.module_utils.network.f5.bigip import HAS_F5SDK
    from ansible.module_utils.network.f5.bigip import F5Client
    from ansible.module_utils.network.f5.common import F5ModuleError
    from ansible.module_utils.network.f5.common import AnsibleF5Parameters
    from ansible.module_utils.network.f5.common import cleanup_tokens
    from ansible.module_utils.network.f5.common import fq_name
    from ansible.module_utils.network.f5.common import f5_argument_spec
    try:
        from ansible.module_utils.network.f5.common import iControlUnexpectedHTTPError
    except ImportError:
        HAS_F5SDK = False


class Parameters(AnsibleF5Parameters):
    api_map = {
        'strictUpdates': 'strict_updates',
        'trafficGroup': 'traffic_group',
    }

    returnables = []

    api_attributes = [
        'tables', 'variables', 'template', 'lists', 'deviceGroup',
        'inheritedDevicegroup', 'inheritedTrafficGroup', 'trafficGroup',
        'strictUpdates'
    ]

    updatables = ['tables', 'variables', 'lists', 'strict_updates', 'traffic_group']

    def to_return(self):
        result = {}
        for returnable in self.returnables:
            result[returnable] = getattr(self, returnable)
        result = self._filter_params(result)
        return result

    @property
    def tables(self):
        result = []
        if not self._values['tables']:
            return None
        tables = self._values['tables']
        for table in tables:
            tmp = dict()
            name = table.get('name', None)
            if name is None:
                raise F5ModuleError(
                    "One of the provided tables does not have a name"
                )
            tmp['name'] = str(name)
            columns = table.get('columnNames', None)
            if columns:
                tmp['columnNames'] = [str(x) for x in columns]
                # You cannot have rows without columns
                rows = table.get('rows', None)
                if rows:
                    tmp['rows'] = []
                    for row in rows:
                        tmp['rows'].append(dict(row=[str(x) for x in row['row']]))
            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    @tables.setter
    def tables(self, value):
        self._values['tables'] = value

    @property
    def variables(self):
        result = []
        if not self._values['variables']:
            return None
        variables = self._values['variables']
        for variable in variables:
            tmp = dict((str(k), str(v)) for k, v in iteritems(variable))
            if 'encrypted' not in tmp:
                # BIG-IP will inject an 'encrypted' key if you don't provide one.
                # If you don't provide one, then we give you the default 'no', by
                # default.
                tmp['encrypted'] = 'no'
            if 'value' not in tmp:
                tmp['value'] = ''

            # This seems to happen only on 12.0.0
            elif tmp['value'] == 'none':
                tmp['value'] = ''
            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    @variables.setter
    def variables(self, value):
        self._values['variables'] = value

    @property
    def lists(self):
        result = []
        if not self._values['lists']:
            return None
        lists = self._values['lists']
        for list in lists:
            tmp = dict((str(k), str(v)) for k, v in iteritems(list) if k != 'value')
            if 'encrypted' not in list:
                # BIG-IP will inject an 'encrypted' key if you don't provide one.
                # If you don't provide one, then we give you the default 'no', by
                # default.
                tmp['encrypted'] = 'no'
            if 'value' in list:
                if len(list['value']) > 0:
                    # BIG-IP removes empty values entries, so mimic this behavior
                    # for user-supplied values.
                    tmp['value'] = [str(x) for x in list['value']]
            result.append(tmp)
        result = sorted(result, key=lambda k: k['name'])
        return result

    @lists.setter
    def lists(self, value):
        self._values['lists'] = value

    @property
    def parameters(self):
        result = dict(
            tables=self.tables,
            variables=self.variables,
            lists=self.lists
        )
        return result

    @parameters.setter
    def parameters(self, value):
        if value is None:
            return
        if 'tables' in value:
            self.tables = value['tables']
        if 'variables' in value:
            self.variables = value['variables']
        if 'lists' in value:
            self.lists = value['lists']
        if 'deviceGroup' in value:
            self.deviceGroup = value['deviceGroup']
        if 'inheritedDevicegroup' in value:
            self.inheritedDevicegroup = value['inheritedDevicegroup']
        if 'inheritedTrafficGroup' in value:
            self.inheritedTrafficGroup = value['inheritedTrafficGroup']
        if 'trafficGroup' in value:
            self.trafficGroup = value['trafficGroup']
        if 'strictUpdates' in value:
            self.strictUpdates = value['strictUpdates']

    @property
    def template(self):
        if self._values['template'] is None:
            return None
        return fq_name(self.partition, self._values['template'])

    @template.setter
    def template(self, value):
        self._values['template'] = value

    @property
    def strict_updates(self):
        if self._values['strict_updates'] is None and self.strictUpdates is None:
            return None

        # Specifying the value overrides any associated value in the payload
        elif self._values['strict_updates'] is True:
            return 'enabled'
        elif self._values['strict_updates'] is False:
            return 'disabled'

        # This will be automatically `None` if it was not set by the
        # `parameters` setter
        elif self.strictUpdates:
            return self.strictUpdates
        else:
            return self._values['strict_updates']

    @property
    def traffic_group(self):
        if self._values['traffic_group'] is None and self.trafficGroup is None:
            return None

        # Specifying the value overrides any associated value in the payload
        elif self._values['traffic_group']:
            result = fq_name(self.partition, self._values['traffic_group'])

        # This will be automatically `None` if it was not set by the
        # `parameters` setter
        elif self.trafficGroup:
            result = fq_name(self.partition, self.trafficGroup)
        else:
            result = fq_name(self.partition, self._values['traffic_group'])
        if result.startswith('/Common/'):
            return result
        else:
            raise F5ModuleError(
                "Traffic groups can only exist in /Common"
            )


class Changes(Parameters):
    pass


class Difference(object):
    def __init__(self, want, have=None):
        self.want = want
        self.have = have

    def compare(self, param):
        try:
            result = getattr(self, param)
            return result
        except AttributeError:
            return self.__default(param)

    def __default(self, param):
        attr1 = getattr(self.want, param)
        try:
            attr2 = getattr(self.have, param)
            if attr1 != attr2:
                return attr1
        except AttributeError:
            return attr1

    @property
    def traffic_group(self):
        if self.want.traffic_group != self.have.traffic_group:
            return self.want.traffic_group


class ModuleManager(object):
    def __init__(self, *args, **kwargs):
        self.module = kwargs.get('module', None)
        self.client = kwargs.get('client', None)
        self.have = None
        self.want = Parameters(params=self.module.params)
        self.changes = Changes()

    def _set_changed_options(self):
        changed = {}
        for key in Parameters.returnables:
            if getattr(self.want, key) is not None:
                changed[key] = getattr(self.want, key)
        if changed:
            self.changes = Changes(params=changed)

    def _update_changed_options(self):
        diff = Difference(self.want, self.have)
        updatables = Parameters.updatables
        changed = dict()
        for k in updatables:
            change = diff.compare(k)
            if change is None:
                continue
            else:
                if isinstance(change, dict):
                    changed.update(change)
                else:
                    changed[k] = change
        if changed:
            self.changes = Changes(params=changed)
            return True
        return False

    def exec_module(self):
        changed = False
        result = dict()
        state = self.want.state

        try:
            if state == "present":
                changed = self.present()
            elif state == "absent":
                changed = self.absent()
        except iControlUnexpectedHTTPError as e:
            raise F5ModuleError(str(e))

        changes = self.changes.to_return()
        result.update(**changes)
        result.update(dict(changed=changed))
        return result

    def exists(self):
        result = self.client.api.tm.sys.application.services.service.exists(
            name=self.want.name,
            partition=self.want.partition
        )
        return result

    def present(self):
        if self.exists():
            return self.update()
        else:
            return self.create()

    def create(self):
        self._set_changed_options()
        if self.want.traffic_group is None and self.want.trafficGroup is None:
            self.want.update({'traffic_group': '/Common/traffic-group-1'})
        if self.module.check_mode:
            return True
        self.create_on_device()
        return True

    def update(self):
        self.have = self.read_current_from_device()
        if not self.should_update() and not self.want.force:
            return False
        if self.module.check_mode:
            return True
        self.update_on_device()
        return True

    def should_update(self):
        result = self._update_changed_options()
        if result:
            return True
        return False

    def update_on_device(self):
        params = self.want.api_params()
        params['execute-action'] = 'definition'
        resource = self.client.api.tm.sys.application.services.service.load(
            name=self.want.name,
            partition=self.want.partition
        )
        resource.update(**params)

    def read_current_from_device(self):
        result = self.client.api.tm.sys.application.services.service.load(
            name=self.want.name,
            partition=self.want.partition
        ).to_dict()
        result.pop('_meta_data', None)
        return Parameters(params=result)

    def create_on_device(self):
        params = self.want.api_params()
        self.client.api.tm.sys.application.services.service.create(
            name=self.want.name,
            partition=self.want.partition,
            **params
        )

    def absent(self):
        if self.exists():
            return self.remove()
        return False

    def remove(self):
        if self.module.check_mode:
            return True
        self.remove_from_device()
        if self.exists():
            raise F5ModuleError("Failed to delete the iApp service")
        return True

    def remove_from_device(self):
        resource = self.client.api.tm.sys.application.services.service.load(
            name=self.want.name,
            partition=self.want.partition
        )
        if resource:
            resource.delete()


class ArgumentSpec(object):
    def __init__(self):
        self.supports_check_mode = True
        argument_spec = dict(
            name=dict(required=True),
            template=dict(),
            parameters=dict(
                type='dict'
            ),
            state=dict(
                default='present',
                choices=['absent', 'present']
            ),
            force=dict(
                default='no',
                type='bool'
            ),
            strict_updates=dict(
                type='bool',
                default='yes'
            ),
            traffic_group=dict(),
            partition=dict(
                default='Common',
                fallback=(env_fallback, ['F5_PARTITION'])
            )
        )
        self.argument_spec = {}
        self.argument_spec.update(f5_argument_spec)
        self.argument_spec.update(argument_spec)


def main():
    spec = ArgumentSpec()

    module = AnsibleModule(
        argument_spec=spec.argument_spec,
        supports_check_mode=spec.supports_check_mode
    )
    if not HAS_F5SDK:
        module.fail_json(msg="The python f5-sdk module is required")

    try:
        client = F5Client(**module.params)
        mm = ModuleManager(module=module, client=client)
        results = mm.exec_module()
        cleanup_tokens(client)
        module.exit_json(**results)
    except F5ModuleError as e:
        cleanup_tokens(client)
        module.fail_json(msg=str(e))


if __name__ == '__main__':
    main()
